/* globals Ext, dss, DSS, _, formatDelayString, rgb2hex, hsv2rgb, addDataIdToElement, secondsToHHMM, secondsToHHMMSS, secondsToHours, secondsToHourMinutes */

const weekdaysCount = {
    NONE: 0,
    SINGLE: 1,
    WEEKEND: 2,
    WORKWEEK: 5,
    ALL: 7
};

Ext.override(Ext.tip.Tip, {
    layout: 'fit'
});

Ext.QuickTips.init();

Ext.parseDevice = function (device, error) {
    if (device) {
        device.deviceName = device.name || '?';
        device.zoneName = device.ZoneID != null ? dss.staticDataModel.getNameForZone(device.ZoneID) : '?';
        device.imageCell = device.present != null
            ? '<td class="icon-small"' + (error ? ' style="' + error + '"' : '') + '><img src="' + dss.staticDataModel.getImageForDevice(device) + '">'
            : '';
    }
    return device;
};

/**
 * Creates a table row for the condition to be used in tooltip
 * @param sStateName    the description of the condition
 * @param bStatus       the current status
 * @returns {String}    returns a HTML formated table row "condition description : status" with translated status string.
 */
function createTooltipEntry (sStateName, bStatus) {
    var sTooltip = '';
    var sStatus = bStatus ? _('fulfilled') : _('not fulfilled');
    sTooltip = '<tr><td>' + sStateName + ':&nbsp;</td><td>' + sStatus + '</td></tr>';
    return sTooltip;
}

/**
 * calculates the current status of the given time frame condition
 * @param stateObject
 * @returns {Boolean}
 */
function checkTimeFrameConditions (stateObject) {
    var oTimeStartNode = stateObject.start;
    var oTimeEndNode = stateObject.end;
    var oNow = new Date();
    var nowSinceMidnight = oNow.getHours() * dss.commonConsts.SECONDS_PER_HOUR + oNow.getMinutes() * dss.commonConsts.SECONDS_PER_MINUTE + oNow.getSeconds();

    if (oTimeStartNode == null || oTimeEndNode == null) {
        return true;
    }

    if (oTimeStartNode.timeBase !== 'daily' || oTimeEndNode.timeBase !== 'daily') {
        return false;
    }

    var startSinceMidnight = oTimeStartNode.offset;
    var endSinceMidnight = oTimeEndNode.offset;

    if (startSinceMidnight < endSinceMidnight) {
        // 1. 10:00:00 - 12:00:00
        if (nowSinceMidnight < startSinceMidnight || nowSinceMidnight > endSinceMidnight) {
            return false;
        }
    } else {
        // 2. 22:00:00 - 04:00:00
        if (nowSinceMidnight < startSinceMidnight && nowSinceMidnight > endSinceMidnight) {
            return false;
        }
    }

    return true;
}

/**
 * calculates the current status of the given week days condition.
 *
 * @param Array of weekdays
 * @returns {Boolean}
 */
function checkWeekdaysConditions (weekDayArray) {
    function dayOfWeekAsInteger (day) {
        return DSS.Weekdays.weekStartingFromSU().indexOf(day);
    }
    var i;
    var oNow = new Date();
    var currentCondState = false;
    if (weekDayArray) {
        if (weekDayArray.length === weekdaysCount.ALL) {
            // daily so its true
            currentCondState = true;
        } else {
            for (i = 0; i < weekDayArray.length; i++) {
                if (dayOfWeekAsInteger(weekDayArray[i]) == oNow.getDay()) {
                    currentCondState = true;
                }
            }
        }
    }
    return currentCondState;
}

/**
 * check if object is a window handle
 *
 * @param system state object
 * @returns {Boolean}
 */
function isObjectWindowHandle (systemStateObject) {
    var oArray = systemStateObject.name.split('.');
    var oDevice;
    var bIsWindowHandle = false;
    if (oArray.length === 3) {
        if (oArray[0] === 'dev') {
            oDevice = dss.staticDataModel.getDevice(oArray[1]);
        }
        if (oDevice != null) {
            bIsWindowHandle = dss.staticDataModel.IsWindowHandle(oDevice);
        }
    }
    return bIsWindowHandle;
}

/**
 * evaluate if condition for window handle is valid
 *
 * @param system state object
 * @returns {Boolean}
 */
function evaluateWindowHandleCondition (systemStateObject) {
    var oArray = systemStateObject.name.split('.');
    if (oArray.length === 3) {
        var oDevice;
        if (oArray[0] === 'dev') {
            oDevice = dss.staticDataModel.getDevice(oArray[1]);
        }

        var index = -1;
        if (oDevice != null) {
            index = dss.staticDataModel.GetWindowHandleBinaryInputIndex(oDevice);
        }

        // check valid return
        if (index < 0) {
            return false;
        }

        var objName = oArray[0] + '.' + oArray[1] + '.' + index;
        var stateArray = dss.staticDataModel.getStateArray();
        var object = stateArray[objName];

        var tmpState = systemStateObject.value == object.value;
        return tmpState;
    }
    return false;
}

/**
 * calculates the current status of the given system state conditions.
 *
 * @param Array of weekdays
 * @returns {Boolean}
 */
function checkSystemStateConditions (systemStateObject) {

    if (isObjectWindowHandle(systemStateObject)) {
        return evaluateWindowHandleCondition(systemStateObject);
    }

    var stateArray = dss.staticDataModel.getStateArray();
    if (stateArray[systemStateObject.name] == null) {
        return false;
    }
    var tmpState = systemStateObject.value.toString() === stateArray[systemStateObject.name].value.toString() ||
        stateArray[systemStateObject.name].state != null && systemStateObject.value.toString() === stateArray[systemStateObject.name].state.toString();
    return tmpState;
}

/**
 * calculates the current status of the given zone state condition
 * @param oZone zone object
 * @returns {Boolean}
 */
function checkZoneStateConditions (oZone) {
    var testZoneId;
    var testGroupId;
    var testSceneId;
    var tmpState = true;
    var sResult;
    var resultObj;
    var lastCalledSceneId;

    testZoneId = oZone.zone;
    testGroupId = oZone.group;
    testSceneId = oZone.scene;

    // check condition state
    sResult = dss.ajaxSyncRequest('/json/property/getInteger', {
        path: '/apartment/zones/zone' + testZoneId + '/groups/group' + testGroupId + '/lastCalledScene'
    });

    resultObj = Ext.JSON.decode(sResult);
    if (resultObj.ok) {
        lastCalledSceneId = resultObj.result.value;
        if (lastCalledSceneId != testSceneId) {
            tmpState = false;
        }
    } else {
        // invalid response
        tmpState = false;
    }

    if (!tmpState) {
        // check condition state again
        sResult = dss.ajaxSyncRequest('/json/property/getInteger', {
            path: '/apartment/zones/zone' + testZoneId + '/groups/group' + testGroupId + '/lastCalledBasicScene'
        });

        resultObj = Ext.JSON.decode(sResult);
        if (resultObj.ok) {
            lastCalledSceneId = resultObj.result.value;
            if (lastCalledSceneId == testSceneId) {
                tmpState = true;
            }
        } else {
            // invalid response
        }
    }
    return tmpState;
}

function checkAddonStateConditions (addonStateObject) {
    var tmpState = true;
    if (addonStateObject.scriptID === 'system-addon-user-defined-states') {
        if (addonStateObject.value == 2) {
            if (dss.staticDataModel.getUserStateValue(addonStateObject.name)) {
                tmpState = false;
            }
        }
        if (addonStateObject.value == 1) {
            if (!dss.staticDataModel.getUserStateValue(addonStateObject.name)) {
                tmpState = false;
            }
        }
    }

    if (addonStateObject.scriptID === 'system-addon-heating-controller') {
        var apartmentControllerState = dss.ajaxSyncRequest('/json/property/query2', {
            query: '/usr/addon-states/system-addon-heating-controller/apartment_controller(value,state)'
        });
        apartmentControllerState = JSON.parse(apartmentControllerState).result.apartment_controller;
        tmpState = apartmentControllerState.value === addonStateObject.value;
    }

    return tmpState;
}

function viewUpRendererOutput (val, metadata, record, rowIdx, colIdx, store, view) {
    if (view.up().customColumnRenderer && view.up().customColumnRenderer[colIdx]) {
        return view.up().customColumnRenderer[colIdx](val, metadata, record, rowIdx, colIdx, store, view);
    }
}

Ext.define('DSS.component.conditionCcolumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.conditioncolumn',
    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        var sString;
        var imageSrc;
        var sTooltip = '';
        var overallCondState = true;
        var tmpStatus;
        var statusString;
        var sTmpString;
        var i;
        var k;
        // TEMP - sText is not used at the moment, but it should be declared and defined
        var sText = '';

        var customRendererOptions = {
            hideEnabled: false,
            hideWeekdays: false,
            hideGenericSystemStates: false,
            hideZoneStates: false,
            hideTimeframe: false,
            filterSystemState: [],
            stateTitles: []
        };

        if (view.up().customRendererOptions) {
            if (view.up().customRendererOptions[colIdx] != null) {
                customRendererOptions = view.up().customRendererOptions[colIdx];
            }
        }
        var oRecord = val;
        var numberOfRows = 0;
        var sResult = ''; // = "<table>";
        if (!customRendererOptions.hideEnabled) {
            if (oRecord.enabled != null) {
                numberOfRows++;
                sResult += '<tr><td colspan="2">' + _('entry is ' + (oRecord.enabled ? 'enabled' : 'disabled')) + '</td></tr>';
            }
        }

        if (!customRendererOptions.hideTimeframe) {
            if (oRecord.timeframe) {
                function calcTimeframeString (oTimeObjectStart, oTimeObjectEnd) {
                    var sStart = '';
                    if (oTimeObjectStart.timeBase === 'daily') {
                        sStart = secondsToHHMM(oTimeObjectStart.offset) + ' ' + _('o\'clock');
                        if (oTimeObjectStart.random != null) {
                            sStart = '&plusmn;' + sText;
                        }
                    } else {
                        sStart = secondsToHHMM(oTimeObjectStart.offset);

                        if ([ 'dusk', 'dawn', 'sunrise', 'sunset'].indexOf(oTimeObjectStart.timeBase) > -1) {
                            if (oTimeObjectStart.offset == 0) {
                                sStart = _('at ' + oTimeObjectStart.timeBase);
                            } else if (oTimeObjectStart.offset < 0) {
                                sStart = Ext.String.format2(_('{time} before ' + oTimeObjectStart.timeBase), { time: sText });
                            } else {
                                sStart = Ext.String.format2(_('{time} after ' + oTimeObjectStart.timeBase), { time: sText });
                            }
                        }
                        if (oTimeObjectStart.limitBefore != null) {
                            sStart += Ext.String.format2(_('<br>but not past {hour}:{minute} o\'clock'), {
                                hour: secondsToHours(oTimeObjectStart.limitBefore),
                                minute: secondsToHourMinutes(oTimeObjectStart.limitBefore)
                            });
                        }
                        if (oTimeObjectStart.limitAfter != null) {
                            sStart += Ext.String.format2(_('<br>but not before {hour}:{minute} o\'clock'), {
                                hour: secondsToHours(oTimeObjectStart.limitAfter),
                                minute: secondsToHourMinutes(oTimeObjectStart.limitAfter)
                            });
                        }
                        if (oTimeObjectStart.random != null) {
                            sStart = '&plusmn;' + sText;
                        }
                    }

                    var sEnd = '';
                    if (oTimeObjectEnd.timeBase === 'daily') {
                        sEnd = secondsToHHMM(oTimeObjectEnd.offset) + ' ' + _('o\'clock');
                        if (oTimeObjectEnd.random != null) {
                            sEnd = '&plusmn;' + sText;
                        }
                    } else {
                        sEnd = secondsToHHMM(oTimeObjectEnd.offset);

                        if ([ 'dusk', 'dawn', 'sunrise', 'sunset'].indexOf(oTimeObjectEnd.timeBase) > -1) {
                            if (oTimeObjectEnd.offset == 0) {
                                sEnd = _('at ' + oTimeObjectEnd.timeBase);
                            } else if (oTimeObjectEnd.offset < 0) {
                                sEnd = Ext.String.format2(_('{time} before ' + oTimeObjectEnd.timeBase), { time: sText });
                            } else {
                                sEnd = Ext.String.format2(_('{time} after ' + oTimeObjectEnd.timeBase), { time: sText });
                            }
                        }
                        if (oTimeObjectEnd.limitBefore != null) {
                            sEnd += Ext.String.format2(_('<br>but not past {hour}:{minute} o\'clock'), {
                                hour: secondsToHours(oTimeObjectEnd.limitBefore),
                                minute: secondsToHourMinutes(oTimeObjectEnd.limitBefore)
                            });
                        }
                        if (oTimeObjectEnd.limitAfter != null) {
                            sEnd += Ext.String.format2(_('<br>but not before {hour}:{minute} o\'clock'), {
                                hour: secondsToHours(oTimeObjectEnd.limitAfter),
                                minute: secondsToHourMinutes(oTimeObjectEnd.limitAfter)
                            });
                        }
                        if (oTimeObjectEnd.random != null) {
                            sEnd = '&plusmn;' + sText;
                        }
                    }

                    return Ext.String.format2(_('only between {startTime} and {endTime}'), {
                        startTime: sStart,
                        endTime: sEnd
                    });
                }
                for (i = 0; i < oRecord.timeframe.length; i++) {
                    numberOfRows++;
                    tmpStatus = checkTimeFrameConditions(oRecord.timeframe[i]);
                    statusString = calcTimeframeString(oRecord.timeframe[i].start, oRecord.timeframe[i].end);
                    sTooltip += createTooltipEntry(statusString, tmpStatus);
                    overallCondState = overallCondState && tmpStatus;
                    sResult += '<tr>' +
                        '<td class="icon-small"><img src="' + dss.staticDataModel.baseImagePath + 'dss/time_condition.png"></td>' +
                        '<td>' + statusString + '</td></tr>';
                }

            }
        }
        if (!customRendererOptions.hideDate) {
            if (oRecord.date != null) {
                for (i = 0; i < oRecord.date.length; i++) {
                    var startDate = oRecord.date[i].start;
                    var startDay = parseInt(startDate.substr(6, 2), 10);
                    var startMonthNumber = parseInt(startDate.substr(4, 2), 10);
                    var startMonth = Ext.Date.monthNumberToMonthName(parseInt(startDate.substr(4, 2), 10));

                    var endDate = oRecord.date[i].end;
                    var endDay = parseInt(endDate.substr(6, 2), 10);
                    var endMonthNumber = parseInt(endDate.substr(4, 2), 10);
                    var endMonth = Ext.Date.monthNumberToMonthName(parseInt(endDate.substr(4, 2), 10));

                    numberOfRows++;
                    sResult += '<tr>' +
                        '<td class="icon-small"><img src="' + dss.staticDataModel.baseImagePath + 'dss/date_condition.png"></td>' +
                        '<td>' + _('Dateframe from') + ' ' + startDay + '. ' + startMonth + ' ' + _('to') + ' ' + endDay + '. ' + endMonth + '</td>' +
                        '</tr>';
                    var actualDate = new Date();
                    var oStartDate = new Date(actualDate.getYear() + 1900, startMonthNumber - 1, startDay);
                    var oEndDate = new Date(actualDate.getYear() + 1900, endMonthNumber - 1, endDay);
                    oEndDate.setHours(23, 59, 59, 999);

                    if (oStartDate > oEndDate) {
                        var oStartDateBefore = new Date(actualDate.getYear() + 1900 - 1, startMonthNumber - 1, startDay);
                        var oEndDateAfter = new Date(actualDate.getYear() + 1900 + 1, endMonthNumber - 1, endDay);
                        tmpStatus = false;
                        if (oStartDateBefore < actualDate && actualDate < oEndDate) {
                            tmpStatus = true;
                        }
                        if (oStartDate < actualDate && actualDate < oEndDateAfter) {
                            tmpStatus = true;
                        }
                    } else {
                        tmpStatus = false;
                        if (oStartDate < actualDate && actualDate < oEndDate) {
                            tmpStatus = true;
                        }
                    }

                    statusString = _('Dateframe');
                    sTooltip += createTooltipEntry(statusString, tmpStatus);

                    overallCondState = overallCondState && tmpStatus;
                }
            }
        }

        if (!customRendererOptions.hideWeekdays) {
            if (oRecord.weekdays != null) {
                function calcWeekdayString (recObject) {
                    if (!recObject) {
                        return '';
                    }
                    if (recObject.length === weekdaysCount.NONE) {
                        return sString + _('no time defined');
                    }
                    if (recObject.length === weekdaysCount.ALL) {
                        return sString + _('daily');
                    }
                    if (recObject.length === weekdaysCount.SINGLE) {
                        return sString + Ext.Date.weekdayToWeekdayDescription(recObject[0]);
                    }

                    const days = recObject.reduce(function (allDays, day) {
                        allDays[day] = true;
                        return allDays;
                    }, {});

                    sString = _('only ');
                    if (recObject.length === weekdaysCount.WEEKEND && days.SA && days.SU) {
                        return sString + _('on weekends');
                    }
                    if (recObject.length === weekdaysCount.WORKWEEK && days.MO && days.TU && days.WE && days.TH && days.FR) {
                        return sString + _('MO-FR');
                    }

                    const selectedDays = [];
                    DSS.Weekdays.weekStartingFromMO().forEach(function (day) {
                        selectedDays.push(days[day]
                            ? '<b>' + _(day) + '</b>'
                            : '<del>' + _(day) + '</del>');
                    });
                    return selectedDays.join(', ');
                }

                numberOfRows++;
                tmpStatus = checkWeekdaysConditions(oRecord.weekdays);
                statusString = calcWeekdayString(oRecord.weekdays);
                sTooltip += createTooltipEntry(statusString, tmpStatus);
                overallCondState = overallCondState && tmpStatus;
                sResult += '<tr>' +
                    '<td class="icon-small"><img src="' + dss.staticDataModel.baseImagePath + 'dss/weekday_condition.png"></td>' +
                    '<td><span>' + statusString + '</span></td>' +
                    '</tr>';
            }
        }

        if (oRecord.systemState) {
            for (i = 0; i < oRecord.systemState.length; i++) {
                tmpStatus = checkSystemStateConditions(oRecord.systemState[i]);
                var fShow = true;
                if (customRendererOptions.hideGenericSystemStates) {
                    fShow = false;
                    if (customRendererOptions.filterSystemState) {
                        for (k = 0; k < customRendererOptions.filterSystemState.length; k++) {
                            if (customRendererOptions.filterSystemState[k] == oRecord.systemState[i].name) {
                                fShow = true;
                            }
                        }
                    }
                }
                var sIcon = dss.staticDataModel.getImagePathForState(oRecord.systemState[i].name, oRecord.systemState[i].value);
                var fShown = false;
                if (fShow) {
                    if (customRendererOptions.stateTitles) {
                        for (k = 0; k < customRendererOptions.stateTitles.length; k++) {
                            var sStateName = oRecord.systemState[i].name;
                            var sSuffix = '';
                            if (sStateName.split('.').length === 2 && dss.staticDataModel.userGroups) {
                                for (var l = 0; l < dss.staticDataModel.userGroups.length; l++) {
                                    if (sStateName.split('.')[1] === 'group' + dss.staticDataModel.userGroups[l].id) {
                                        sStateName = sStateName.split('.')[0];
                                        sSuffix = ' ' + _('in') + ' ' + dss.staticDataModel.userGroups[l].name;
                                    }
                                }
                            }
                            if (customRendererOptions.stateTitles[k].name == sStateName) {
                                if (customRendererOptions.stateTitles[k].value == oRecord.systemState[i].value) {
                                    if (sIcon != null) {
                                        numberOfRows++;
                                        sResult += '<tr>' +
                                            '<td class="icon-small" colspan="1"><img src="' + sIcon + '"></td>' +
                                            '<td>' + customRendererOptions.stateTitles[k].title + sSuffix + '</td>' +
                                            '</tr>';
                                    } else {
                                        numberOfRows++;
                                        sResult += '<tr><td colspan="2">' + customRendererOptions.stateTitles[k].title + sSuffix + '</td></tr>';
                                    }
                                    sTooltip += createTooltipEntry(customRendererOptions.stateTitles[k].title + sSuffix, tmpStatus);
                                    overallCondState = overallCondState && tmpStatus;
                                    fShown = true;
                                }
                            }
                        }
                    }
                }
                if (!fShown) {
                    var sError = '';
                    var oArray = oRecord.systemState[i].name.split('.');
                    if (oArray.length === 3) {
                        if (oArray[0] === 'dev') {
                            var oDevice = dss.staticDataModel.getDevice(oArray[1]);
                            if (oDevice != null) {
                                sString = (oDevice.name ? oDevice.name : '?') + ' ' + _('in') + ' ';
                                sString += oDevice.ZoneID != null ? dss.staticDataModel.getNameForZone(oDevice.ZoneID) : '?';

                                var oParameter = {
                                    device: oDevice.name,
                                    zone_of_device: dss.staticDataModel.getNameForZone(oDevice.ZoneID)
                                };

                                var fSpecialSkip = false;
                                if (oDevice.specialActions) {
                                    if (oDevice.specialActions.conditions) {
                                        if (oDevice.specialActions.conditions.akmEvent) {
                                            fSpecialSkip = true;
                                            for (var sKey in oDevice.specialActions.conditions.akmEvent) {
                                                if (oDevice.specialActions.conditions.akmEvent[sKey].id == oArray[2]) {
                                                    if (oDevice.specialActions.conditions.akmEvent[sKey].value == oRecord.systemState[i].value) {
                                                        oParameter.state_value = oDevice.specialActions.conditions.akmEvent[sKey].name;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                if (!fSpecialSkip) {
                                    if (oDevice.deviceStates) {
                                        if (oDevice.deviceStates[oArray[2]] != null) {
                                            var oState = oDevice.deviceStates[oArray[2]];
                                            var title = oState.title;
                                            var values = oState.options;
                                            fSpecialSkip = true;
                                            if (values[oRecord.systemState[i].value] != null) {
                                                oParameter.state_value = title + ': ' + values[oRecord.systemState[i].value];
                                            } else {
                                                oParameter.state_value = title + ': ' + oRecord.systemState[i].value;
                                            }
                                        }
                                    }
                                }
                                if (!fSpecialSkip) {
                                    var typeIndex = parseInt(oArray[2], 10);
                                    var type = 255;
                                    if (oDevice.binaryInputs) {
                                        for (k = 0; k < oDevice.binaryInputs.length; k++) {
                                            if (oDevice.binaryInputs[k].index == typeIndex) {
                                                type = oDevice.binaryInputs[k].type;
                                            }
                                        }
                                    }

                                    var bIsWindowHandle = dss.staticDataModel.IsWindowHandle(oDevice);
                                    if (bIsWindowHandle) {
                                        oParameter.state_value = dss.staticDataModel.getNameForAKMValue(type, oRecord.systemState[i].value);
                                    } else if (oRecord.systemState[i].value == 1) {
                                        oParameter.state_value = dss.staticDataModel.getNameForAKMValue(type, true);
                                    } else {
                                        oParameter.state_value = dss.staticDataModel.getNameForAKMValue(type, false);
                                    }
                                }
                                sString = Ext.String.format2(_('{#grid_condition_column_akm_state}{state_value} on {device} in {zone_of_device}'), oParameter);

                                numberOfRows++;
                                sTooltip += createTooltipEntry(sString, tmpStatus);
                                overallCondState = overallCondState && tmpStatus;
                                imageSrc = oDevice.ZoneID != null ? dss.staticDataModel.getImageForDevice(oDevice) : '';
                                imageSrc = imageSrc ? '<td class="icon-small"><img src="' + imageSrc + '"></td>' : '';
                                sResult += '<tr>' + imageSrc + '<td ' + (imageSrc ? '' : 'colspan="2" ') + 'style="' + sError + '">' + sString + '</td></tr>';
                                continue;
                            }
                        }

                        if (oArray[0] === 'zone') {
                            var sNameOfZone = dss.staticDataModel.getNameForZone(parseInt(oArray[1], 10));
                            if (oArray[2] === 'light') {
                                if (oRecord.systemState[i].value == 1) {
                                    numberOfRows++;
                                    sString = Ext.String.format2(_('{#grid_condition_column_zone_state}light in {zone} is on'), {
                                        zone: sNameOfZone
                                    });
                                    sResult += '<tr>' +
                                        '<td class="icon-small">' +
                                        '<img src="' + dss.staticDataModel.getImagePathForScene(DSS.Zones.ALL_DEVICES, DSS.Groups.LIGHT, DSS.Scenes.USER1) + '">' +
                                        '</td>' +
                                        '<td style="' + sError + '">' + sString + '</td>' +
                                        '</tr>';
                                } else {
                                    numberOfRows++;
                                    sString = Ext.String.format2(_('{#grid_condition_column_zone_state}light in {zone} is off'), {
                                        zone: sNameOfZone
                                    });
                                    sResult += '<tr>' +
                                        '<td class="icon-small">' +
                                        '<img src="' + dss.staticDataModel.getImagePathForScene(DSS.Zones.ALL_DEVICES, DSS.Groups.LIGHT, DSS.Scenes.OFF) + '">' +
                                        '</td>' +
                                        '<td style="' + sError + '">' + sString + '</td>' +
                                        '</tr>';
                                }
                                sTooltip += createTooltipEntry(sString, tmpStatus);
                                overallCondState = overallCondState && tmpStatus;
                                continue;
                            } else
                            if (oArray[2] === 'motion') {
                                if (oRecord.systemState[i].value == 1) {
                                    numberOfRows++;
                                    sString = Ext.String.format2(_('{#grid_condition_column_zone_state}motion in {zone} is reported'), {
                                        zone: sNameOfZone
                                    });
                                    sResult += '<tr>' +
                                        '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForAKMValue(5, true) + '"></td>' +
                                        '<td style="' + sError + '">' + sString + '</td>' +
                                        '</tr>';
                                } else {
                                    numberOfRows++;
                                    sString = Ext.String.format2(_('{#grid_condition_column_zone_state}no motion in {zone} is reported'), {
                                        zone: sNameOfZone
                                    });
                                    sResult += '<tr>' +
                                        '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForAKMValue(5, false) + '"></td>' +
                                        '<td style="' + sError + '">' + sString + '</td>' +
                                        '</tr>';
                                }
                                sTooltip += createTooltipEntry(sString, tmpStatus);
                                overallCondState = overallCondState && tmpStatus;
                                continue;
                            } else
                            if (oArray[2] === 'presence') {
                                if (oRecord.systemState[i].value == 1) {
                                    numberOfRows++;
                                    sString = Ext.String.format2(_('{#grid_condition_column_zone_state}presence in {zone} is reported'), {
                                        zone: sNameOfZone
                                    });
                                    sTooltip += createTooltipEntry(sString, tmpStatus);
                                    overallCondState = overallCondState && tmpStatus;
                                    sResult += '<tr>' +
                                        '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForAKMValue(1, true) + '"></td>' +
                                        '<td style="' + sError + '">' + sString + '</td>' +
                                        '</tr>';
                                } else {
                                    numberOfRows++;
                                    sString = Ext.String.format2(_('{#grid_condition_column_zone_state}absence in {zone} is reported'), {
                                        zone: sNameOfZone
                                    });
                                    sTooltip += createTooltipEntry(sString, tmpStatus);
                                    overallCondState = overallCondState && tmpStatus;
                                    sResult += '<tr>' +
                                        '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForAKMValue(1, false) + '"></td>' +
                                        '<td style="' + sError + '">' + sString + '</td>' +
                                        '</tr>';
                                }
                                continue;
                            } else
                            if (oArray[2] === 'heating') {
                                if (oRecord.systemState[i].value == 1) {
                                    numberOfRows++;
                                    sString = Ext.String.format2(_('{#grid_condition_column_zone_state}heating in {zone} is activ'), {
                                        zone: sNameOfZone
                                    });
                                    sTooltip += createTooltipEntry(sString, tmpStatus);
                                    overallCondState = overallCondState && tmpStatus;
                                    sResult += '<tr>' +
                                        '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForAKMValue(1, true) + '"></td>' +
                                        '<td style="' + sError + '">' + sString + '</td>' +
                                        '</tr>';
                                } else {
                                    numberOfRows++;
                                    sString = Ext.String.format2(_('{#grid_condition_column_zone_state}heating in {zone} is inactiv'), {
                                        zone: sNameOfZone
                                    });
                                    sTooltip += createTooltipEntry(sString, tmpStatus);
                                    overallCondState = overallCondState && tmpStatus;
                                    sResult += '<tr>' +
                                        '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForAKMValue(1, false) + '"></td>' +
                                        '<td style="' + sError + '">' + sString + '</td>' +
                                        '</tr>';
                                }
                                continue;
                            }
                        }

                        if (oArray[0] === 'cluster') {
                            var sNameOfCluster = dss.staticDataModel.getUserGroupName(parseInt(oArray[1], 10));
                            if (oArray[2] === 'operation_lock') {
                                if (oRecord.systemState[i].value == 1) {
                                    numberOfRows++;
                                    sString = Ext.String.format2(_('{#grid_condition_column_cluster_state}operation in {cluster} is locked'), {
                                        cluster: sNameOfCluster
                                    });
                                    sResult += '<tr>' +
                                        '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForState('operation_lock', 'active') + '"></td>' +
                                        '<td style="' + sError + '">' + sString + '</td>' +
                                        '</tr>';
                                } else {
                                    numberOfRows++;
                                    sString = Ext.String.format2(_('{#grid_condition_column_cluster_state}operation in {cluster} is free'), {
                                        cluster: sNameOfCluster
                                    });
                                    sResult += '<tr>' +
                                        '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForState('operation_lock', 'inactive') + '"></td>' +
                                        '<td style="' + sError + '">' + sString + '</td>' +
                                        '</tr>';
                                }
                                sTooltip += createTooltipEntry(sString, tmpStatus);
                                overallCondState = overallCondState && tmpStatus;
                                continue;
                            }
                        }
                    } else if (oArray.length === 6) {
                        if (oArray[0] === 'zone' && oArray[2] === 'group' && oArray[4] === 'status') {
                            var zoneId = parseInt(oArray[1], 10);
                            var groupId = parseInt(oArray[3], 10);
                            var groupName = dss.staticDataModel.getNameForGroup(zoneId, groupId);
                            var stateName = oArray[5];
                            var icon = dss.staticDataModel.getImagePathForState(oRecord.systemState[i].name, oRecord.systemState[i].value);
                            var active = oRecord.systemState[i].value === 'active' || oRecord.systemState[i].value == '1';
                            if (zoneId === DSS.Zones.ALL_DEVICES) {
                                if (groupId === DSS.Groups.APARTMENT_VENTILATION) {
                                    // this is global application state - apartment ventilation
                                    if (stateName === 'malfunction') {
                                        sString = (active ? 'Malfunction' : 'no Malfunction');
                                    } else if (stateName === 'service') {
                                        sString = (active ? 'Service needed' : 'no Service needed');
                                    }
                                    sString = Ext.String.format2(_('{#editor_condition_column_zone_state}' + sString + ' in {group}'), {
                                        group: groupName
                                    });
                                    if (!dss.staticDataModel.globalApplicationsAvaible) {   // apartment ventilation is not present in the system (no such devices in zone 0)
                                        sError = 'color:#ff0000;';
                                    }
                                }
                            } else {
                                // this is zone state
                                var sNameOfZone = dss.staticDataModel.getNameForZone(zoneId);
                                if (stateName === 'malfunction') {
                                    sString = (active ? 'Malfunction detected' : 'no Malfunction');
                                } else if (stateName === 'service') {
                                    sString = (active ? 'Service needed' : 'no Service needed');
                                }
                                sString = Ext.String.format2(_('{#editor_condition_column_zone_state}' + sString + ' in {zone} for {group}'), {
                                    zone: sNameOfZone,
                                    group: groupName
                                });
                            }
                            numberOfRows++;
                            sResult += '<tr><td class="icon-small"><img src="' + icon + '"></td><td style="' + sError + '">' + sString + '</td></tr>';
                            sTooltip += createTooltipEntry(sString, tmpStatus);
                            overallCondState = overallCondState && tmpStatus;
                            continue;
                        }
                    }

                    const stateDescription = Ext.String.format2(
                        _('{#grid_condition_column_generic_state}' + dss.staticDataModel.getNameForState(oRecord.systemState[i].name, oRecord.systemState[i].value)),
                        {}
                    );
                    if (sIcon != null) {
                        numberOfRows++;
                        sTooltip += createTooltipEntry(stateDescription, tmpStatus);
                        overallCondState = overallCondState && tmpStatus;
                        sResult += '<tr>' +
                            '<td class="icon-small" colspan="1"><img src="' + sIcon + '"></td>' +
                            '<td>' + stateDescription + '</td>' +
                            '</tr>';
                    } else {
                        numberOfRows++;
                        sTooltip += createTooltipEntry(stateDescription, tmpStatus);
                        overallCondState = overallCondState && tmpStatus;
                        sResult += '<tr>' +
                            '<td colspan="2">' + stateDescription + '</td>' +
                            '</tr>';
                    }
                }
            }
        }
        if (!customRendererOptions.hideZoneStates) {
            if (oRecord.zoneState != null) {
                for (i = 0; i < oRecord.zoneState.length; i++) {
                    numberOfRows++;
                    var oZone = oRecord.zoneState[i];
                    tmpStatus = checkZoneStateConditions(oZone);
                    sResult += '<tr><td class="icon-small">' + dss.staticDataModel.getImageForScene(oZone.zone, oZone.group, oZone.scene) + '</td>';

                    sResult += '<td>';
                    sTmpString = dss.staticDataModel.getNameForScene(oZone.zone, oZone.group, oZone.scene);
                    if (oZone.group != DSS.Groups.ALL) {
                        sTmpString += ' ' + _('for') + ' ' + dss.staticDataModel.getNameForGroup(oZone.zone, oZone.group);
                    }
                    if (oZone.zone != DSS.Zones.ALL_DEVICES) {
                        sTmpString += ' ' + _('in') + ' ' + dss.staticDataModel.getNameForZone(oZone.zone);
                    } else if (oZone.scene < DSS.Scenes.ZONE_SCENES_END_AT && oZone.group < DSS.Groups.APARTMENT_VENTILATION) {
                        sTmpString += ' ' + _('in all rooms');
                    }
                    sResult += sTmpString;
                    sResult += '</td></tr>';
                    sTooltip += createTooltipEntry(sTmpString, tmpStatus);
                    overallCondState = overallCondState && tmpStatus;
                }
            }
        }

        if (oRecord.addonStates != null) {
            for (i = 0; i < oRecord.addonStates.length; i++) {
                numberOfRows++;
                sResult += '<tr>';
                sTmpString = '';
                var textString = '';
                var image;
                tmpStatus = checkAddonStateConditions(oRecord.addonStates[i]);

                if (oRecord.addonStates[i].scriptID === 'system-addon-user-defined-states') {
                    if (dss.staticDataModel.userStates) {
                        for (var j = 0; j < dss.staticDataModel.userStates.length; j++) {
                            if (dss.staticDataModel.userStates[j].name == oRecord.addonStates[i].name) {
                                var valueText;
                                if (oRecord.addonStates[i].value == 1) {
                                    valueText = dss.staticDataModel.userStates[j].setName;
                                    image = dss.staticDataModel.baseImagePath + 'state/uds_state_active.png';
                                } else {
                                    valueText = dss.staticDataModel.userStates[j].resetName;
                                    image = dss.staticDataModel.baseImagePath + 'state/uds_state_inactive.png';
                                }
                                textString = Ext.String.format2(_('Customstate {state} is {value}'), {
                                    state: dss.staticDataModel.userStates[j].displayName,
                                    value: valueText
                                });
                                sTmpString = '<td class="icon-small"><img src="' + image + '"></td><td>' + textString;
                            }
                        }
                    }
                }

                if (oRecord.addonStates[i].scriptID === 'system-addon-heating-controller') {
                    if (oRecord.addonStates[i].name === 'apartment_controller') {
                        switch (parseInt(oRecord.addonStates[i].value, 10)) {
                            case 0:
                                textString = 'grid_climate_off';
                                image = dss.staticDataModel.baseImagePath + 'state/heating_mode_off.png';
                                break;
                            case 1:
                                textString = 'grid_climate_heating';
                                image = dss.staticDataModel.baseImagePath + 'state/heating_mode_heating.png';
                                break;
                            case 2:
                                textString = 'grid_climate_cooling';
                                image = dss.staticDataModel.baseImagePath + 'state/heating_mode_cooling.png';
                                break;
                        }
                        textString = _(textString);
                        sTmpString = '<td class="icon-small"><img src="' + image + '"></td><td>' + textString;
                    }
                }

                if (!sTmpString) {
                    textString = oRecord.addonStates[i].name;
                    textString += ' ' + _('must be') + ' ';
                    textString += oRecord.addonStates[i].value;
                    textString += ' ' + _('for App') + ' ';
                    textString += oRecord.addonStates[i].scriptID;
                    sTmpString = '<td colspan="2" style="color:#ff0000;">' + textString;
                }
                sResult += sTmpString + '</td></tr>';
                sTooltip += createTooltipEntry(Ext.String.htmlEncode(unescape(textString)), tmpStatus);
                overallCondState = overallCondState && tmpStatus;
            }
        }
        if (numberOfRows > 0) {
            // add a title and the table tags
            sTooltip = '<table><tr><th>' + _('Current status of conditions') + '</th></tr>' + sTooltip + '</table>';
            // insert the condition state image cell
            var nIndex = sResult.indexOf('<tr>');
            sResult = sResult.slice(nIndex + '<tr>'.length);
            sResult = '<table><tr><td rowspan="' + numberOfRows + '"><img src="' + dss.staticDataModel.getImagePathForConditionState(overallCondState) + '">&nbsp</td>' + sResult;
        } else {
            sResult = '<table>' + sResult;
        }
        var fActive = view.up().isFieldActive(record);
        metadata.tdAttr = 'data-qtip="' + sTooltip + '"';
        if (!fActive) {
            return '<span style="color:grey;">' + sResult + '</table>' + '</span>';
        } else {
            return sResult + '</table>';
        }
    }
});

Ext.define('DSS.component.triggerColumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.triggercolumn',
    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        var sError = '';
        var oRecord = val;
        var sResult = '';
        var f1tip = false;
        var f2tip = false;
        var f3tip = false;
        var sString;
        var idString;
        var oDevice;
        var valueString;
        var oUserStates;
        var oParameter;
        var i;
        var j;
        var sKey;
        var fSpecial;
        var fSkip;

        for (i = 0; oRecord && i < oRecord.length; i++) {
            var oAction = oRecord[i];

            var fValid = dss.validateTriggerObject(oAction);
            if (fValid == dss.validValue.completlyWrong) {
                sError = 'color:#ff0000;';
            }
            if (fValid == dss.validValue.temporaryWrong) {
                sError = 'color:#aaaaaa;';
            }
            if (fValid == dss.validValue.ok) {
                sError = 'color:#000000;';
            }

            sResult += '<tr>';
            switch (oAction.type) {
                case 'addon-state-change': {
                    if (oAction['addon-id'] === 'system-addon-user-defined-states') {
                        idString = oAction.name;
                        valueString = oAction.state;
                        oUserStates = dss.staticDataModel.getUserStates();
                        oParameter = {};
                        var image;
                        if (oUserStates) {
                            for (j = 0; j < oUserStates.length; j++) {
                                if (oUserStates[j].name == idString) {
                                    oParameter.state = oUserStates[j].displayName;
                                    if (oAction.state === 'active') {
                                        oParameter.value = oUserStates[j].setName;
                                        image = dss.staticDataModel.baseImagePath + 'state/uds_state_active.png';
                                    } else {
                                        oParameter.value = oUserStates[j].resetName;
                                        image = dss.staticDataModel.baseImagePath + 'state/uds_state_inactive.png';
                                    }
                                }
                            }
                        }

                        sString = '<td class="pad-right h18" style="' + sError + '"></td>';
                        if (image != null) {
                            sString = '<td class="icon-small" style="' + sError + '"><img src="' + image + '"></td>';
                        }
                        sString += '<td style="' + sError + '">' + Ext.String.format2(_('Customstate {state} is {value}'), oParameter);
                        sString += '</td>';
                        sResult += sString;
                        break;
                    // } else if (oAction['addon-id'] === 'system-addon-heating-controller') {
                    //     idString = oAction['name'];
                    //     valueString = oAction['state'];
                    //
                    //     var image = null;
                    //     if ([ 'off', 'heating', 'cooling' ].indexOf(valueString) > -1) {
                    //         image = dss.staticDataModel.baseImagePath + 'state/heating_mode_' + valueString + '.png';
                    //     }
                    //
                    //     var sString;
                    //     if (image != null) {
                    //         sString = '<td class="icon-small" style="' + sError + '"><img src="' + image + '"></td>';
                    //     } else {
                    //         sString = '<td class="icon-small" style="' + sError + '"></td>';
                    //     }
                    //
                    //     if (idString === 'apartment_controller') {
                    //         sString += '<td style="' + sError + '">' + _('heating mode ' + valueString);
                    //     } else {
                    //         sString += '<td style="' + sError + '"><b>' + _(idString) + '</b> ' + _('is') + ' <b>' + _(valueString) + '</b>';
                    //     }
                    //     sString += '</td>';
                    //     sResult += sString;
                    //     break;
                    } else {
                        sString = '<td class="pad-right h18" style="' + sError + '"></td>';
                        sString += '<td style="' + sError + '"><b>' + oAction.name + '</b> change to <b>' + oAction.state + '</b> ' +
                            'for App <b>' + oAction['addon-id'] + '</b> (internal)';
                        sString += '</td>';
                        sResult += sString;
                    }
                    break;
                }

                case 'event': {
                    switch (oAction.name) {
                        case 'sunshine': {
                            var stateStr = oAction.parameter.value == 1 ? 'active' : 'inactive';
                            sString = '<td class="icon-small" style="' + sError + '">' +
                                '<img src="' + dss.staticDataModel.getImagePathForState('sun', stateStr) + '">' +
                                '</td>';
                            sString += '<td style="' + sError + '">' +
                                Ext.String.format2(_('Sun protection {direction} is {state}'), {
                                    direction: oAction.parameter.direction,
                                    state: stateStr
                                }) +
                                '</td>';
                            sResult += sString;
                            break;
                        }
                    }
                    break;
                }

                case 'device-named-event': {
                    fSkip = false;
                    idString = oAction.id;

                    oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(oAction.dsuid), sError);
                    if (oDevice != null) {
                        sString = oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName;
                        if (oDevice.deviceEvents) {
                            if (oDevice.deviceEvents[idString]) {
                                sResult += oDevice.imageCell +
                                    '<td style="' + sError + '">' + oDevice.deviceEvents[idString].title + ' ' + _('from') + ' ' + sString + '</td>';
                                fSkip = true;
                            }
                        }
                    }
                    if (!fSkip) {
                        sResult += '<td colspan="2">' + idString + _('from') + oAction.dsuid + '</td>';
                    }
                    break;
                }

                case 'device-named-action': {
                    idString = oAction.id;

                    oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(oAction.dsuid), sError);
                    if (oDevice) {
                        sString = oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName;

                        if (oDevice.deviceStandardActions) {
                            if (oDevice.deviceStandardActions[idString]) {
                                sResult += oDevice.imageCell +
                                    '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                    oDevice.deviceStandardActions[idString].title + ' ' + _('from') + ' ' + sString +
                                    '</td>';
                                fSkip = true;
                            }
                        }
                        if (oDevice.deviceActions) {
                            if (oDevice.deviceActions[idString]) {
                                sResult += oDevice.imageCell +
                                    '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                    oDevice.deviceActions[idString].title + ' ' + _('from') + ' ' + sString +
                                    '</td>';
                                fSkip = true;
                            }
                        }
                        if (oDevice.deviceCustomActions) {
                            if (oDevice.deviceCustomActions[idString]) {
                                sResult += oDevice.imageCell +
                                    '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                    oDevice.deviceCustomActions[idString].title + ' ' + _('from') + ' ' + sString +
                                    '</td>';
                                fSkip = true;
                            }
                        }
                    }
                    if (!fSkip) {
                        sResult += '<td colspan="2">' + idString + _('from') + oAction.dsuid + '</td>';
                    }
                    break;
                }

                case 'state-change': {
                    idString = oAction.name;
                    valueString = oAction.state;
                    oUserStates = dss.staticDataModel.getUserStates();
                    oParameter = {};
                    var fSkipAfterUserstateFound = false;
                    if (oUserStates) {
                        for (j = 0; j < oUserStates.length; j++) {
                            if (oUserStates[j].name == idString) {
                                oParameter.state = oUserStates[j].displayName;
                                if (oAction.state === 'active') {
                                    oParameter.value = oUserStates[j].setName;
                                } else {
                                    oParameter.value = oUserStates[j].resetName;
                                }
                                sResult += '<td colspan="2" style="' + sError + '">' +
                                    Ext.String.format2(_('Customstate {state} is {value}'), oParameter) +
                                    '</td>';
                                fSkipAfterUserstateFound = true;
                            }
                        }
                    }
                    if (fSkipAfterUserstateFound) {
                        break;
                    }

                    var oArray = oAction.name.split('.');
                    if (oArray.length === 3) {
                        if (oArray[0] === 'dev') {
                            oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(oArray[1]), sError);
                            if (oDevice.present != null) {
                                fSkip = false;
                                if ((oDevice.hasStates) && (dss.staticDataModel.getDeviceStates(oArray[1]))) {
                                    oParameter = {
                                        device: oDevice.name,
                                        zone_of_device: dss.staticDataModel.getNameForZone(oDevice.ZoneID)
                                    };
                                    var states = dss.staticDataModel.getDeviceStates(oArray[1]);
                                    for (sKey in states) {
                                        if (sKey == oArray[2]) {
                                            oParameter.state_name = states[sKey].title;
                                            for (var sSubkey in states[sKey].options) {
                                                if (sSubkey == oAction.state) {
                                                    oParameter.state_value = states[sKey].options[sSubkey];
                                                    sString = Ext.String.format2(
                                                        _('{#grid_trigger_column_vdc_state}{state_value} for {state_name} on {device} in {zone_of_device}'),
                                                        oParameter
                                                    );
                                                    sResult += oDevice.imageCell +
                                                        '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                                        sString +
                                                        '</td>';
                                                    fSkip = true;
                                                    break;
                                                }
                                            }
                                        }
                                    }
                                    if (fSkip) {
                                        break;
                                    }
                                } else {
                                    var typeIndex = parseInt(oArray[2], 10);
                                    oParameter = {
                                        device: oDevice.name,
                                        zone_of_device: dss.staticDataModel.getNameForZone(oDevice.ZoneID)
                                    };

                                    if (oDevice.specialActions && oDevice.specialActions.triggers && oDevice.specialActions.triggers.binaryInputs) {
                                        for (sKey in oDevice.specialActions.triggers.binaryInputs) {
                                            const binaryInput = oDevice.specialActions.triggers.binaryInputs[sKey];
                                            if (binaryInput.id.toString() == typeIndex.toString()) {
                                                if (oAction.state === 'active') {
                                                    if (binaryInput.value === 'active') {
                                                        fSkip = true;
                                                        oParameter.state_value = binaryInput.name;
                                                    }
                                                }
                                                if (oAction.state === 'inactive') {
                                                    if (binaryInput.value === 'inactive') {
                                                        fSkip = true;
                                                        oParameter.state_value = binaryInput.name;
                                                    }
                                                }
                                            }
                                        }
                                    }

                                    if (!fSkip) {
                                        var bIsWindowHandle = dss.staticDataModel.IsWindowHandle(oDevice);
                                        var type = 255;
                                        if (oDevice.binaryInputs) {
                                            for (var k = 0; k < oDevice.binaryInputs.length; k++) {
                                                if (oDevice.binaryInputs[k].index == typeIndex) {
                                                    type = oDevice.binaryInputs[k].type;
                                                }
                                            }
                                        }
                                        if (!bIsWindowHandle) {
                                            oParameter.state_value = dss.staticDataModel.getNameForAKMValue(type, oAction.state === 'active');
                                        } else if (oAction.state === 'closed') {
                                            oParameter.state_value = dss.staticDataModel.getNameForAKMValue(type, dss.whState.WH_Closed);
                                        } else if (oAction.state === 'open') {
                                            oParameter.state_value = dss.staticDataModel.getNameForAKMValue(type, dss.whState.WH_Open);
                                        } else if (oAction.state === 'tilted') {
                                            oParameter.state_value = dss.staticDataModel.getNameForAKMValue(type, dss.whState.WH_Tilted);
                                        }
                                    }
                                    sString = Ext.String.format2(_('{#grid_trigger_column_akm_state}{state_value} on {device} in {zone_of_device}'), oParameter);
                                    sResult += oDevice.imageCell +
                                        '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' + sString + '</td>';
                                    break;
                                }
                            } else {
                                sString = Ext.String.format2(_('{#grid_trigger_column_akm_state}{state_value} on {device} in {zone_of_device}'), {
                                    state_value: '?',
                                    device: '?',
                                    zone_of_device: '?'
                                });
                                sResult += '<td colspan="2" style="' + sError + '">' + sString + '</td>';
                                break;
                            }
                        }
                        if (oArray[0] === 'zone') {
                            var sNameOfZone = dss.staticDataModel.getNameForZone(parseInt(oArray[1], 10));
                            if (oArray[2] === 'light') {
                                if (oAction.state === 'active') {
                                    sString = Ext.String.format2(_('{#grid_trigger_column_zone_state}light in {zone} is on'), {
                                        zone: sNameOfZone
                                    });
                                    sResult += '<td class="icon-small">' +
                                        '<img src="' + dss.staticDataModel.getImagePathForScene(DSS.Zones.ALL_DEVICES, DSS.Groups.LIGHT, DSS.Scenes.USER1) + '">' +
                                        '</td>' +
                                        '<td style="' + sError + '">' + sString + '</td>';
                                } else {
                                    sString = Ext.String.format2(_('{#grid_trigger_column_zone_state}light in {zone} is off'), {
                                        zone: sNameOfZone
                                    });
                                    sResult += '<td class="icon-small">' +
                                        '<img src="' + dss.staticDataModel.getImagePathForScene(DSS.Zones.ALL_DEVICES, DSS.Groups.LIGHT, DSS.Scenes.OFF) + '">' +
                                        '</td>' +
                                        '<td style="' + sError + '">' + sString + '</td>';
                                }
                                break;
                            } else
                            if (oArray[2] === 'motion') {
                                if (oAction.state === 'active') {
                                    sString = Ext.String.format2(_('{#grid_trigger_column_zone_state}motion in {zone} is reported'), {
                                        zone: sNameOfZone
                                    });
                                    sResult += '<td class="icon-small">' +
                                        '<img src="' + dss.staticDataModel.getImagePathForAKMValue(5, true) + '">' +
                                        '</td>' +
                                        '<td style="' + sError + '">' + sString + '</td>';
                                } else {
                                    sString = Ext.String.format2(_('{#grid_trigger_column_zone_state}no motion in {zone} is reported'), {
                                        zone: sNameOfZone
                                    });
                                    sResult += '<td class="icon-small">' +
                                        '<img src="' + dss.staticDataModel.getImagePathForAKMValue(5, false) + '">' +
                                        '</td>' +
                                        '<td style="' + sError + '">' + sString + '</td>';
                                }
                                break;
                            } else
                            if (oArray[2] === 'presence') {
                                if (oAction.state === 'active') {
                                    sString = Ext.String.format2(_('{#grid_trigger_column_zone_state}presence in {zone} is reported'), {
                                        zone: sNameOfZone
                                    });
                                    sResult += '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForAKMValue(1, true) + '"></td>' +
                                        '<td style="' + sError + '">' + sString + '</td>';
                                } else {
                                    sString = Ext.String.format2(_('{#grid_trigger_column_zone_state}absence in {zone} is reported'), {
                                        zone: sNameOfZone
                                    });
                                    sResult += '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForAKMValue(1, false) + '"></td>' +
                                        '<td style="' + sError + '">' + sString + '</td>';
                                }
                                break;
                            }
                        }
                        if (oArray[0] === 'cluster') {
                            if (oArray[2] === 'operation_lock') {
                                if (oAction.state === 'active') {
                                    sString = Ext.String.format2(_('{#grid_trigger_column_cluster_state}Operation in {cluster} is locked'), {
                                        cluster: oArray[1]
                                    });
                                    sResult += '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForState('operation_lock', 'active') + '"></td>' +
                                        '<td style="' + sError + '">' + sString + '</td>';
                                } else {
                                    sString = Ext.String.format2(_('{#grid_trigger_column_cluster_state}Operation in {cluster} is free'), {
                                        cluster: oArray[1]
                                    });
                                    sResult += '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForState('operation_lock', 'inactive') + '"></td>' +
                                        '<td style="' + sError + '">' + sString + '</td>';
                                }
                                break;
                            }
                        }
                    } else if (oArray.length === 2) {
                        if (oArray[0] === 'wind') {
                            var clusters = dss.staticDataModel.getUsergroups();
                            var clusterId = parseInt(oArray[1].substring(5), 10);
                            var clusterName = Ext.String.format2(_('Group {group}'), {
                                group: clusterId
                            });
                            for (var c = 0; c < clusters.length; c++) {
                                var cluster = clusters[c];
                                if (cluster.id === clusterId) {
                                    clusterName = cluster.name;
                                    break;
                                }
                            }
                            if (oAction.state === 'active') {
                                sString = Ext.String.format2(_('{#grid_trigger_column_cluster_state}Wind in {cluster}'), {
                                    cluster: clusterName
                                });
                                sResult += '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForState('wind', 'active') + '"></td>' +
                                    '<td style="' + sError + '">' + sString + '</td>';
                            } else {
                                sString = Ext.String.format2(_('{#grid_trigger_column_cluster_state}No Wind in {cluster}'), {
                                    cluster: clusterName
                                });
                                sResult += '<td class="icon-small"><img src="' + dss.staticDataModel.getImagePathForState('wind', 'inactive') + '"></td>' +
                                    '<td style="' + sError + '">' + sString + '</td>';
                            }
                            break;
                        }
                    } else if (oArray.length === 6) {
                        if (oArray[0] === 'zone' && oArray[2] === 'group' && oArray[4] === 'status') {
                            var zoneId = parseInt(oArray[1], 10);
                            var groupId = parseInt(oArray[3], 10);
                            var groupName = dss.staticDataModel.getNameForGroup(zoneId, groupId);
                            var stateName = oArray[5];
                            var icon = dss.staticDataModel.getImagePathForState(oAction.name, oAction.state);
                            var active = oAction.state === 'active' || oAction.state == '1';
                            if (zoneId === DSS.Zones.ALL_DEVICES) {
                                if (groupId === DSS.Groups.APARTMENT_VENTILATION) {
                                    // this is global application state - apartment ventilation
                                    if (stateName === 'malfunction') {
                                        sString = (active ? 'Malfunction' : 'no Malfunction');
                                    } else if (stateName === 'service') {
                                        sString = (active ? 'Service needed' : 'no Service needed');
                                    }
                                    sString = Ext.String.format2(_('{#editor_trigger_zone_state}' + sString + ' in {group}'), {
                                        group: groupName
                                    });
                                }
                            } else {
                                // this is zone state
                                var sNameOfZone = dss.staticDataModel.getNameForZone(zoneId);
                                if (stateName === 'malfunction') {
                                    sString = (active ? 'Malfunction detected' : 'no Malfunction');
                                } else if (stateName === 'service') {
                                    sString = (active ? 'Service needed' : 'no Service needed');
                                }
                                sString = Ext.String.format2(_('{#editor_trigger_zone_state}' + sString + ' in {zone} for {group}'), {
                                    zone: sNameOfZone,
                                    group: groupName
                                });
                                if (!dss.staticDataModel.getDevicesForZone(zoneId, groupId).length) {   // e.g. room ventilation is not present in the given zone
                                    sError = 'color:#ff0000;';
                                }
                            }
                            sResult += '<td class="icon-small"><img src="' + icon + '"></td><td style="' + sError + '">' + sString + '</td>';
                            break;
                        }
                    }

                    sString = '<td class="icon-small" style="' + sError + '">' +
                        '<img src="' + dss.staticDataModel.getImagePathForState(oAction.name, oAction.state) + '"></td>';
                    sString += '<td style="' + sError + '">' + dss.staticDataModel.getNameForState(oAction.name, oAction.state);
                    sString += '</td>';
                    sResult += sString;
                    break;
                }

                case 'device-sensor': {
                    oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(oAction.dsuid), sError);

                    sString = oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName;
                    var fSpecialString = false;

                    if (oDevice.specialActions.triggers && oDevice.specialActions.triggers.deviceEvents) {
                        for (sKey in oDevice.specialActions.triggers.deviceEvents) {
                            if (oDevice.specialActions.triggers.deviceEvents[sKey].id != null) {
                                if (oDevice.specialActions.triggers.deviceEvents[sKey].id == oAction.eventid) {
                                    fSpecialString = true;
                                }
                            }
                        }
                    }

                    if (!fSpecialString) {
                        sString = _('consumptionmessage') + ' ' + dss.staticDataModel.getNameForSensorEvent(oDevice, oAction.eventid) + ' ' + _('from') + ' ' + sString + '</td>';
                    } else {
                        sString = dss.staticDataModel.getNameForSensorEvent(oDevice, oAction.eventid) + ' ' + _('from') + ' ' + sString + '</td>';
                    }
                    sResult += oDevice.imageCell +
                        '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' + sString;
                    break;
                }

                case 'undo-zone-scene': { // scene,
                    // group,zone,force
                    sString = '<td class="icon-small" style="' + sError + '">' +
                           dss.staticDataModel.getImageForScene(oAction.zone, oAction.group, oAction.scene, true) + '</td>';

                    sString += '<td style="' + sError + '">' + _('Undo ') + dss.staticDataModel.getNameForScene(oAction.zone, oAction.group, oAction.scene);
                    if (oAction.group != 0) {
                        sString += ' ' + _('for') + ' ' + dss.staticDataModel.getNameForGroup(oAction.zone, oAction.group);
                    }
                    if (oAction.zone != 0) {
                        sString += ' ' + _('in') + ' ' + dss.staticDataModel.getNameForZone(oAction.zone);
                    }
                    sString += '</td>';
                    sResult += sString;
                    break;
                }

                case 'bus-zone-scene':
                case 'zone-scene': { // scene,
                    // group,zone,force
                    sString = '<td class="icon-small" style="' + sError + '">' +
                        dss.staticDataModel.getImageForScene(oAction.zone, oAction.group, oAction.scene) +
                        '</td>';

                    sString += '<td style="' + sError + '">';
                    if (oAction.forced) {
                        sString += _('all ');
                    }
                    sString += dss.staticDataModel.getNameForScene(oAction.zone, oAction.group, oAction.scene);

                    if (oAction.group != DSS.Groups.ALL) {
                        sString += ' ' + _('for') + ' ' + dss.staticDataModel.getNameForGroup(oAction.zone, oAction.group);
                    }
                    if (oAction.zone != DSS.Zones.ALL_DEVICES && oAction.group < DSS.Groups.APARTMENT_VENTILATION) {
                        sString += ' ' + _('in') + ' ' + dss.staticDataModel.getNameForZone(oAction.zone);
                    }
                    sString += '</td>';
                    sResult += sString;
                    break;
                }

                case 'device-scene': {
                    oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(oAction.dsuid), sError);

                    sString = oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName;

                    switch (oAction.scene) {
                        case 50:
                            if (oDevice.tasterMode == 0 && oDevice.buttonType == 65) {
                                sResult += oDevice.imageCell +
                                    '<td ' + (oDevice.imageCell ? '' : 'callspan="2" ') + 'style="' + sError + '">' +
                                    _('forced off') + ' ' + _('for') + ' ' + sString +
                                    '</td>';
                            } else {
                                sResult += oDevice.imageCell +
                                    '<td ' + (oDevice.imageCell ? '' : 'callspan="2" ') + 'style="' + sError + '">' +
                                    _('local off from') + ' ' + sString +
                                    '</td>';
                            }
                            break;
                        case 51:
                            if (oDevice.tasterMode == 0 && oDevice.buttonType == 65) {
                                sResult += oDevice.imageCell +
                                    '<td ' + (oDevice.imageCell ? '' : 'callspan="2" ') + 'style="' + sError + '">' +
                                    _('forced on') + ' ' + _('for') + ' ' + sString +
                                    '</td>';
                            } else {
                                sResult += oDevice.imageCell +
                                    '<td ' + (oDevice.imageCell ? '' : 'callspan="2" ') + 'style="' + sError + '">' +
                                    _('local on from') + ' ' + sString +
                                    '</td>';
                            }
                            break;
                        case 15:
                            if (oDevice.tasterMode == 0 && oDevice.buttonType == 65) {
                                sResult += oDevice.imageCell +
                                    '<td ' + (oDevice.imageCell ? '' : 'callspan="2" ') + 'style="' + sError + '">' +
                                    _('controlled mode') + ' ' + _('for') + ' ' + sString +
                                    '</td>';
                            }
                    }
                    break;
                }

                case 'device-msg': { // value, dsuid
                    oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(oAction.dsuid), sError);

                    sString = oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName;
                    fSpecial = false;

                    if (oDevice.specialActions && oDevice.specialActions.triggers && oDevice.specialActions.triggers.deviceMsg) {
                        for (sKey in oDevice.specialActions.triggers.deviceMsg) {
                            if (oDevice.specialActions.triggers.deviceMsg[sKey].id == oAction.msg) {
                                sResult += oDevice.imageCell +
                                    '<td ' + (oDevice.imageCell ? '' : 'callspan="2" ') + 'style="' + sError + '">' +
                                    sString + ': ' + oDevice.specialActions.triggers.deviceMsg[sKey].name +
                                    '</td>';
                                fSpecial = true;
                            }
                        }
                    }
                    if (!fSpecial) {
                        switch (oAction.msg) {
                            case 7:
                            case 0:
                                if (skippedFirst == null) {
                                    if (sError === 'color:#ff0000;') {
                                        var skippedFirst = true;
                                        break;
                                    } else {
                                        var skippedFirst = false;
                                    }
                                } else if (!skippedFirst) {
                                    break;
                                }

                                if (oAction.buttonIndex != null && oAction.msg === 0) {
                                    f1tip = true;
                                    if (oAction.buttonIndex == 2) {
                                        sResult += oDevice.imageCell +
                                            '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                            _('switch down') + ' ' + _('from') + ' ' + sString +
                                            '</td>';
                                    }
                                    if (oAction.buttonIndex == 1) {
                                        sResult += oDevice.imageCell +
                                            '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                            _('switch up') + ' ' + _('from') + ' ' + sString +
                                            '</td>';
                                    }
                                }

                                if (!f1tip) {
                                    sResult += oDevice.imageCell +
                                        '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                        _('1x click') + ' ' + _('from') + ' ' + sString +
                                        '</td>';
                                }
                                f1tip = true;
                                break;
                            case 8:
                            case 1:
                                if (!f2tip) {
                                    sResult += oDevice.imageCell +
                                        '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                        _('2x click') + ' ' + _('from') + ' ' + sString +
                                        '</td>';
                                }
                                f2tip = true;
                                break;
                            case 9:
                            case 2:
                                if (!f3tip) {
                                    sResult += oDevice.imageCell +
                                        '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                        _('3x click') + ' ' + _('from') + ' ' + sString +
                                        '</td>';
                                }
                                f3tip = true;
                                break;
                            case 3:
                                sResult += oDevice.imageCell +
                                    '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                    _('4x click') + ' ' + _('from') + ' ' + sString +
                                    '</td>';
                                break;
                            case 6:
                                sResult += oDevice.imageCell +
                                    '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                    _('hold') + ' ' + _('from') + ' ' + sString +
                                    '</td>';
                                break;
                            case 10:
                                sResult += oDevice.imageCell +
                                    '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                    _('short-long-click') + ' ' + _('from') + ' ' + sString +
                                    '</td>';
                                break;
                        }
                    }
                    break;
                }

                case 'device-action': { // value, dsuid
                    oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(oAction.dsuid), sError);

                    sString = oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName;
                    fSpecial = false;

                    if (oDevice.specialActions && oDevice.specialActions.triggers) {
                        if (oDevice.specialActions.triggers.deviceAction) {
                            for (sKey in oDevice.specialActions.triggers.deviceAction) {
                                if (oDevice.specialActions.triggers.deviceAction[sKey].id == oAction.action) {
                                    sResult += oDevice.imageCell +
                                        '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                        sString + ': ' + oDevice.specialActions.triggers.deviceAction[sKey].name +
                                        '</td>';
                                    fSpecial = true;
                                }
                            }
                        }
                        if (oDevice.specialActions.triggers.deviceActions) {
                            for (sKey in oDevice.specialActions.triggers.deviceActions) {
                                if (oDevice.specialActions.triggers.deviceActions[sKey].action == oAction.action) {
                                    sResult += oDevice.imageCell +
                                        '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                        sString + ': ' + oDevice.specialActions.triggers.deviceActions[sKey].name +
                                        '</td>';
                                    fSpecial = true;
                                }
                            }
                        }
                    }
                    if (!fSpecial) {
                        sResult += oDevice.imageCell +
                            '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                            _('Action') + ' ' + (oAction.action - 16) + ' ' + _('from') + ' ' + sString +
                            '</td>';
                    }
                    break;
                }

                case 'device-binary-input': { // value, dsuid
                    oDevice = oAction.dsuid && Ext.parseDevice(dss.staticDataModel.getDevice(oAction.dsuid), sError) || {
                        deviceName: '?',
                        zoneName: '?',
                        imageCell: ''
                    };

                    sString = oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName;

                    switch (oAction.state.toString()) {
                        case '0':
                            sResult += oDevice.imageCell +
                                '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                _('switch up') + ' ' + _('from') + ' ' + sString +
                                '</td>';
                            break;
                        case '1':
                            sResult += oDevice.imageCell +
                                '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' +
                                _('switch down') + ' ' + _('from') + ' ' + sString +
                                '</td>';
                            break;
                    }
                    break;
                }

                case 'custom-event': { // event
                    sResult += '<td style="' + sError + '" colspan="2">' +
                        _('Action') + ': ' + dss.staticDataModel.getHLENameFromId(oAction.event) +
                        '</td>';
                    break;
                }
            }
            sResult += '</tr>';
        }

        return (view.up().isFieldActive(record)
            ? '<table>'
            : '<table style="color:grey;">') + sResult + '</table>';
    }
});

Ext.define('DSS.component.actionColumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.actioncolumn',
    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        var sError = '';
        var oRecord = val;
        var sResult = '';
        var fShowDelay = false;
        var oAction;
        var i;
        var j;
        var fSpecialString;
        var sKey;
        var sName;

        if (oRecord) {
            for (i = 0; i < oRecord.length; i++) {
                oAction = oRecord[i];
                if (oAction.delay != null) {
                    if (oAction.delay != 0) {
                        fShowDelay = true;
                    }
                }
            }
        }

        if (fShowDelay) {
            oRecord.sort(function (a, b) {
                if (a.delay == null) {
                    a.delay = 0;
                }
                if (b.delay == null) {
                    b.delay = 0;
                }
                return a.delay - b.delay;
            });
        }

        for (i = 0; oRecord && i < oRecord.length; i++) {
            oAction = oRecord[i];

            var fValid = dss.validateActionObject(oAction);
            if (fValid == dss.validValue.completlyWrong) {
                sError = 'color:#ff0000;';
            }
            if (fValid == dss.validValue.temporaryWrong) {
                sError = 'color:#aaaaaa;';
            }

            sResult += '<tr>';
            if (fShowDelay) {
                if (oAction.delay == 0) {
                    sResult += '<td></td>';
                } else {
                    sResult += '<td align="right"> +' + formatDelayString(oAction.delay) + '</td>';
                }
            }
            var sString;
            var oDevice;
            switch (oAction.type) {
                case DSS.ActionTypes.DEVICE_MOC_VALUE: {
                    oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(oAction.dsuid), sError);

                    sString = oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName;

                    j = 0;
                    var mocChannel = {};
                    while (oAction['channelid-' + j] != null) {
                        mocChannel['' + oAction['channelid-' + j]] = oAction['channelvalue-' + j];
                        j++;
                    }
                    if (mocChannel['1'] != null) {
                        if (mocChannel['2'] != null) {
                            if (mocChannel['3'] != null) {
                                var hsv = {
                                    hue: mocChannel['2'] / 255.0,
                                    bright: mocChannel['1'] / 255.0,
                                    sat: mocChannel['3'] / 255.0
                                };
                                var sColor = rgb2hex(hsv2rgb(hsv));
                                sResult += oDevice.imageCell +
                                    '<td class="pad-right h18" ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' + sString + '</td>' +
                                    '<td><div style="width:24px;height:24px;border-radius:12px;border:1px solid #bbbbbb;background-color:' + sColor + ';" /></td>';
                            }
                        }
                    }
                    if (mocChannel['1'] != null) {
                        if (mocChannel['4'] != null) {
                            var temperature = dss.MiredToKelvin(dss.MOCValueToMired(mocChannel['4']));
                            var bright = Math.round(mocChannel['1'] / 2.55);

                            sResult += oDevice.imageCell +
                                '<td class="pad-right h18" ' + (oDevice.imageCell ? '' : 'colspan="2" ') + 'style="' + sError + '">' + sString + '</td>' +
                                '<td class="pad-right h18">' + _('ColTemperature') + ': ' + temperature + ' K ' + _('Brightness') + ': ' + bright + ' %</td>' +
                                '<td></td>';
                        }
                    }
                    break;
                }

                case DSS.ActionTypes.DEVICE_ACTION: {
                    oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(oAction.dsuid), sError);

                    sString = oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName;
                    var sTitle = oAction.id;

                    var actions = dss.staticDataModel.getStandardActions(oAction.dsuid);
                    for (sKey in actions) {
                        if (sKey == oAction.id) {
                            sTitle = actions[sKey].title;
                        }
                    }
                    actions = dss.staticDataModel.getDeviceActions(oAction.dsuid);
                    for (sKey in actions) {
                        if (sKey == oAction.id) {
                            sTitle = actions[sKey].title;
                        }
                    }
                    actions = dss.staticDataModel.getCustomActions(oAction.dsuid);
                    for (sKey in actions) {
                        if (sKey == oAction.id) {
                            sTitle = actions[sKey].title;
                        }
                    }

                    sResult += oDevice.imageCell +
                        '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + ' class="h18" style="' + sError + '">' + sString + ': ' + sTitle + '</td>';
                    break;
                }

                case DSS.ActionTypes.CHANGE_ADDON_STATE: {
                    if (oAction['addon-id'] === 'system-addon-user-defined-states') {
                        var stateName = oAction.statename;
                        var value = oAction.state;
                        var oUserStates = dss.staticDataModel.getUserStates();
                        var sStringValue;
                        var image;
                        if (oUserStates) {
                            for (j = 0; j < oUserStates.length; j++) {
                                if (oUserStates[j].name == stateName) {
                                    var valueText;
                                    if (value === 'active') {
                                        valueText = oUserStates[j].setName;
                                        image = dss.staticDataModel.baseImagePath + 'state/uds_state_active.png';
                                    } else {
                                        valueText = oUserStates[j].resetName;
                                        image = dss.staticDataModel.baseImagePath + 'state/uds_state_inactive.png';
                                    }
                                    sStringValue = Ext.String.format2(_('Customstate {state} change to {value}'), {
                                        state: oUserStates[j].displayName,
                                        value: valueText
                                    });
                                }
                            }
                        }

                        sString = '<td class="pad-right h18" style="' + sError + '"></td>';
                        if (image != null) {
                            sString = '<td class="icon-small" style="' + sError + '"><img src="' + image + '"></td>';
                        }
                        sString += '<td style="' + sError + '">' + sStringValue;
                        sString += '</td>';
                        sResult += sString;
                        break;
                    // } else if (oAction['addon-id'] === 'system-addon-heating-controller') {
                    //     var stateName = oAction.statename;
                    //     var value = oAction.state;
                    //
                    //     var image = null;
                    //     if ([ 'off', 'heating', 'cooling' ].indexOf(value) > -1) {
                    //         image = dss.staticDataModel.baseImagePath + 'state/heating_mode_' + value + '.png';
                    //     }
                    //
                    //     var sString;
                    //     if (image != null) {
                    //         sString = '<td class="icon-small" style="' + sError + '"><img src="' + image + '"></td>';
                    //     } else {
                    //         sString += '<td class="icon-small" style="' + sError + '"></td>';
                    //     }
                    //
                    //     if (stateName === 'apartment_controller') {
                    //         sString += '<td style="' + sError + '">' + _('heating mode ' + value);
                    //     } else {
                    //         sString += '<td style="' + sError + '"><b>' + _(stateName) + '</b> ' + _('change to') + ' <b>' + value + '</b>';
                    //     }
                    //     sString += '</td>';
                    //     sResult += sString;
                    //     break;
                    } else {
                        sString = '<td class="pad-right h18" style="' + sError + '"></td>';
                        sString += '<td style="' + sError + '"><b>' + oAction.statename + '</b> change to <b>' + oAction.state + '</b> ' +
                            'for App <b>' + oAction['addon-id'] + '</b> (internal)';
                        sString += '</td>';
                        sResult += sString;
                    }
                    break;
                }

                case DSS.ActionTypes.CHANGE_STATE: {
                    var splitState = oAction.statename.split('.');
                    sString = '<td class="icon-small" style="' + sError + '">' +
                        '<img src="' + dss.staticDataModel.getImagePathForState(splitState[0], oAction.state) + '"></td>';
                    sString += '<td style="' + sError + '">' + dss.staticDataModel.getNameForState(splitState[0], oAction.state);
                    if (splitState.length > 1) {
                        var clusters = dss.staticDataModel.getUsergroups();
                        var clusterId = parseInt(splitState[1].substring(5), 10);
                        var clusterName = Ext.String.format2(_('Group {group}'), {
                            group: clusterId
                        });
                        for (var c = 0; c < clusters.length; c++) {
                            var cluster = clusters[c];
                            if (cluster.id === clusterId) {
                                clusterName = cluster.name;
                                break;
                            }
                        }
                        sString += ' ' + _('in') + ' ' + clusterName;
                    }
                    sString += '</td>';
                    sResult += sString;
                    break;
                }

                case DSS.ActionTypes.ZONE_VALUE: { // scene, group,zone,force
                    sString = '<td colspan="2" class="h18" style="' + sError + '">' + Math.round(oAction.value / 2.55) + '%';
                    if (oAction.group != DSS.Groups.ALL) {
                        sString += ' ' + _('for') + ' ' + dss.staticDataModel.getNameForGroup(oAction.zone, oAction.group);
                    }
                    if (oAction.zone != DSS.Zones.ALL_DEVICES) {
                        sString += ' ' + _('in') + ' ' + dss.staticDataModel.getNameForZone(oAction.zone);
                    } else if (oAction.group < DSS.Groups.APARTMENT_VENTILATION) {
                        sString += ' ' + _('in all rooms');
                    }

                    sString += '</td>';
                    sResult += sString;
                    break;
                }

                case DSS.ActionTypes.HEATING_MODE: {
                    sString = '<table><tr><td class="icon-small" style="' + sError + '">' +
                        dss.staticDataModel.getImageForScene(oAction.zone, DSS.Groups.TEMPERATURE, oAction.mode) +
                        '</td>';
                    sString += '<td style="' + sError + '">';

                    switch (oAction.mode) {
                        case 1:
                            sString += _('COMFORT Mode');
                            break;
                        case 2:
                            sString += _('ECO Mode');
                            break;
                        case 3:
                            sString += oAction.reset ? _('reset NotUsed Mode') : _('NotUsed Mode');
                            break;
                        case 4:
                            sString += oAction.reset ? _('reset NIGHT Mode') : _('NIGHT Mode');
                            break;
                        case 5:
                            sString += oAction.reset ? _('reset HOLIDAY Mode') : _('HOLIDAY Mode');
                            break;
                        case 6:
                            sString += oAction.reset ? _('reset COOLING Mode') : _('COOLING Mode');
                            break;
                        default:
                            sString += '-';
                            break;
                    }

                    sString += ' ' + _('for') + ' ';
                    sString += _('Heating Modes');

                    if (oAction.zone != DSS.Zones.ALL_DEVICES) {
                        sString += ' ' + _('in') + ' ' + dss.staticDataModel.getNameForZone(oAction.zone);
                    }

                    sString += '</td>';
                    sResult += sString;
                    break;
                }

                case DSS.ActionTypes.UNDO_ZONE_SCENE: { // scene,
                    // group,zone,force
                    sString = '<td class="icon-small">' +
                        dss.staticDataModel.getImageForScene(oAction.zone, oAction.group, oAction.scene, true) +
                        '</td>';

                    sString += '<td>' + _('Undo ') + dss.staticDataModel.getNameForScene(oAction.zone, oAction.group, oAction.scene);
                    if (oAction.group != DSS.Groups.ALL) {
                        sString += ' ' + _('for') + ' ' + dss.staticDataModel.getNameForGroup(oAction.zone, oAction.group);
                    }
                    if (oAction.zone != DSS.Zones.ALL_DEVICES) {
                        sString += ' ' + _('in') + ' ' + dss.staticDataModel.getNameForZone(oAction.zone);
                    }
                    if (oAction.force) {
                        sString += ' !';
                    }
                    sString += '</td>';
                    sResult += sString;
                    break;
                }

                case DSS.ActionTypes.ZONE_SCENE: { // scene, group,zone,force
                    sString = '<td class="icon-small" style="' + sError + '">' +
                        dss.staticDataModel.getImageForScene(oAction.zone, oAction.group, oAction.scene) +
                        '</td>';

                    sString += '<td style="' + sError + '">';
                    if (oAction.force) {
                        sString += _('all ');
                    }
                    sString += dss.staticDataModel.getNameForScene(oAction.zone, oAction.group, oAction.scene);
                    if (oAction.group != DSS.Groups.ALL) {
                        sString += ' ' + _('for') + ' ' + dss.staticDataModel.getNameForGroup(oAction.zone, oAction.group);
                    }
                    if (oAction.zone != DSS.Zones.ALL_DEVICES) {
                        sString += ' ' + _('in') + ' ' + dss.staticDataModel.getNameForZone(oAction.zone);
                    } else if (dss.staticDataModel.isClusterGroup(oAction.group)) {
                        // usergruppen haben einen eigenen namen
                    } else if (oAction.scene < DSS.Scenes.ZONE_SCENES_END_AT && oAction.group < DSS.Groups.APARTMENT_VENTILATION) {
                        sString += ' ' + _('in all rooms');
                    }
                    sString += '</td>';
                    sResult += sString;
                    break;
                }
                case DSS.ActionTypes.DEVICE_SCENE: { // scene, dsuid, force
                    oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(oAction.dsuid), sError);
                    if (fValid != dss.validValue.completlyWrong) {
                        if (
                            oAction.scene == DSS.Scenes.IMPULSE && (
                                !dss.staticDataModel.deviceCanDoImpulse(oDevice) ||
                                dss.staticDataModel.deviceHasTwoStages(oDevice) ||
                                dss.staticDataModel.deviceHasThreeStages(oDevice)
                            )
                        ) {
                            sError = 'color:#ff0000;';
                        }
                    }

                    sString = oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName;

                    fSpecialString = false;
                    if (oDevice.specialActions && oDevice.specialActions.actions && oDevice.specialActions.actions.deviceActions) {
                        for (sKey in oDevice.specialActions.actions.deviceActions) {
                            sName = oDevice.specialActions.actions.deviceActions[sKey].name;
                            if (oDevice.specialActions.actions.deviceActions[sKey].scene != null) {
                                if (oDevice.specialActions.actions.deviceActions[sKey].scene == oAction.scene) {
                                    sResult += oDevice.imageCell +
                                        '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + ' class="h18" style="' + sError + '">' +
                                        sString + ' ' + sName +
                                        '</td>';
                                    fSpecialString = true;
                                }
                            }
                        }
                    }

                    if (!fSpecialString) {
                        var sAction;
                        if (
                            (
                                dss.staticDataModel.deviceCanDoImpulse(oDevice) ||
                                dss.staticDataModel.deviceHasTwoStages(oDevice) ||
                                dss.staticDataModel.deviceHasThreeStages(oDevice)
                            ) && [ DSS.Scenes.IMPULSE, DSS.Scenes.MIN, DSS.Scenes.MAX ].includes(oAction.scene)
                        ) {
                            if (oDevice.GroupID === DSS.Groups.JOKER) {
                                if (oAction.scene == DSS.Scenes.IMPULSE) {
                                    sString = '<td class="icon-small">' +
                                        '<img src="' + dss.staticDataModel.baseImagePath + 'sceneIcons/scene_impulse.png' + '">' +
                                        '</td>' +
                                        '<td style="' + sError + '">' + sString + ' ' + _('Impulse') + '</td>';
                                } else if (oAction.scene == DSS.Scenes.MIN) {
                                    sString = '<td class="icon-small">' +
                                        '<img src="' + dss.staticDataModel.baseImagePath + 'sceneIcons/scene_impulse_off.png' + '">' +
                                        '</td>' +
                                        '<td style="' + sError + '">' + sString + ' ' + _('turn off') + '</td>';
                                } else if (oAction.scene == DSS.Scenes.MAX) {
                                    sString = '<td class="icon-small">' +
                                    '<img src="' + dss.staticDataModel.baseImagePath + 'sceneIcons/scene_impulse_on.png' + '">' +
                                    '</td>' +
                                    '<td style="' + sError + '">' + sString + ' ' + _('turn on') + '</td>';
                                }
                            } else {
                                if (oAction.scene == DSS.Scenes.MIN) {
                                    sString = oDevice.imageCell + '<td style="' + sError + '">' + sString + ' ' + _('turn off') + '</td>';
                                } else if (oAction.scene == DSS.Scenes.MAX) {
                                    sString = oDevice.imageCell + '<td style="' + sError + '">' + sString + ' ' + _('turn on') + '</td>';
                                } else if (oAction.scene == DSS.Scenes.IMPULSE) {
                                    sString = oDevice.imageCell + '<td style="' + sError + '">' + sString + ' ' + _('Impulse') + '</td>';
                                }
                            }
                        } else if (oAction.scene == DSS.Scenes.MIN) {
                            sAction = oDevice.GroupID == DSS.Groups.SHADOW ? ' ' + _('close') : ' ' + _('turn off');
                            sString = oDevice.imageCell +
                                '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + ' class="h18" style="' + sError + '">' +
                                sString + sAction +
                                '</td>';
                        } else if (oAction.scene == DSS.Scenes.MAX) {
                            sAction = oDevice.GroupID == DSS.Groups.SHADOW ? ' ' + _('open') : ' ' + _('turn on');
                            sString = oDevice.imageCell +
                                '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + ' class="h18" style="' + sError + '">' +
                                sString + sAction +
                                '</td>';
                        } else if (oAction.scene == DSS.Scenes.LOCAL_ON) {
                            sString = oDevice.imageCell +
                                '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + ' class="h18" style="' + sError + '">' +
                                sString + _(' turn local on') +
                                '</td>';
                        } else {
                            sString = dss.staticDataModel.getNameForScene(oDevice.ZoneID, oDevice.GroupID, oAction.scene) + ' ' + _('for') + ' ' + sString + '</td>';
                            if (oAction.force) {
                                sString = _('all ') + sString;
                            }
                            const sceneImage = dss.staticDataModel.getImageForScene(oDevice.ZoneID, oDevice.GroupID, oAction.scene);
                            sString = (sceneImage
                                ? '<td class="icon-small" style="' + sError + '">' + sceneImage + '</td>'
                                : oDevice.imageCell
                            ) +
                                '<td style="' + sError + '">' + sString + '</td>';
                        }
                        sResult += sString;
                    }
                    break;
                }

                case DSS.ActionTypes.DEVICE_VALUE: { // value, dsuid
                    oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(oAction.dsuid), sError);
                    if (fValid != dss.validValue.completlyWrong) {
                        if (
                            dss.staticDataModel.deviceCanDoImpulse(oDevice) ||
                            dss.staticDataModel.deviceHasTwoStages(oDevice) && (oAction.value === 96 || oAction.value == 160) ||
                            dss.staticDataModel.deviceHasThreeStages(oDevice) && oAction.value === 128
                        ) {
                            sError = 'color:#ff0000;';
                        }
                    }

                    fSpecialString = false;
                    if (oDevice.specialActions.actions && oDevice.specialActions.actions.deviceActions) {
                        for (sKey in oDevice.specialActions.actions.deviceActions) {
                            sName = oDevice.specialActions.actions.deviceActions[sKey].name;
                            if (oDevice.specialActions.actions.deviceActions[sKey].outputValue != null) {
                                if (oDevice.specialActions.actions.deviceActions[sKey].outputValue == oAction.value) {
                                    sResult += oDevice.imageCell +
                                        '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + ' class="h18" style="' + sError + '">' +
                                        oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName + ' ' + sName +
                                        '</td>';
                                    fSpecialString = true;
                                }
                            }
                        }
                    }

                    if (!fSpecialString) {
                        if (dss.staticDataModel.deviceHasTwoStages(oDevice) || dss.staticDataModel.deviceHasThreeStages(oDevice)) {
                            sString = oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName;
                            if (oDevice.GroupID === DSS.Groups.JOKER) {
                                if (oAction.value == 0) { // aus
                                    sString = '<td class="icon-small" style="' + sError + '">' +
                                        '<img src="' + dss.staticDataModel.baseImagePath + 'sceneIcons/scene_impulse_off.png' + '">' + '</td>' +
                                        '<td style="' + sError + '">' + sString + ' ' + _('turn off') + '</td>';
                                }
                                if (oAction.value == 96) { // aus
                                    sString = '<td class="icon-small" style="' + sError + '">' +
                                        '<img src="' + dss.staticDataModel.baseImagePath + 'sceneIcons/scene_impulse_25.png' + '">' + '</td>' +
                                        '<td style="' + sError + '">' + sString + ' ' + _('set to Stage 1') + '</td>';
                                }
                                if (oAction.value == 128) { // aus
                                    sString = '<td class="icon-small" style="' + sError + '">' +
                                        '<img src="' + dss.staticDataModel.baseImagePath + 'sceneIcons/scene_impulse_33.png' + '">' + '</td>' +
                                        '<td style="' + sError + '">' + sString + ' ' + _('set to Stage 1') + '</td>';
                                }
                                if (oAction.value == 160) { // aus
                                    sString = '<td class="icon-small" style="' + sError + '">' +
                                        '<img src="' + dss.staticDataModel.baseImagePath + 'sceneIcons/scene_impulse_50.png' + '">' + '</td>' +
                                        '<td style="' + sError + '">' + sString + ' ' + _('set to Stage 2') + '</td>';
                                }
                                if (oAction.value == 255 && dss.staticDataModel.deviceHasTwoStages(oDevice)) { // aus
                                    sString = '<td class="icon-small" style="' + sError + '">' +
                                        '<img src="' + dss.staticDataModel.baseImagePath + 'sceneIcons/scene_impulse_66.png' + '">' + '</td>' +
                                        '<td style="' + sError + '">' + sString + ' ' + _('set to Stage 2') + '</td>';
                                }
                                if (oAction.value == 255 && dss.staticDataModel.deviceHasThreeStages(oDevice)) { // aus
                                    sString = '<td class="icon-small" style="' + sError + '">' +
                                        '<img src="' + dss.staticDataModel.baseImagePath + 'sceneIcons/scene_impulse_75.png' + '">' + '</td>' +
                                        '<td style="' + sError + '">' + sString + ' ' + _('set to Stage 3') + '</td>';
                                }
                            } else {
                                if (oAction.value == 0) { // aus
                                    sString = oDevice.imageCell +
                                    '<td style="' + sError + '">' + sString + ' ' + _('turn off') + '</td>';
                                }
                                if (oAction.value == 96) { // aus
                                    sString = oDevice.imageCell +
                                        '<td style="' + sError + '">' + sString + ' ' + _('set to Stage 1') + '</td>';
                                }
                                if (oAction.value == 128) { // aus
                                    sString = oDevice.imageCell +
                                        '<td style="' + sError + '">' + sString + ' ' + _('set to Stage 1') + '</td>';
                                }
                                if (oAction.value == 160) { // aus
                                    sString = oDevice.imageCell +
                                        '<td style="' + sError + '">' + sString + ' ' + _('set to Stage 2') + '</td>';
                                }
                                if (oAction.value == 255 && dss.staticDataModel.deviceHasTwoStages(oDevice)) { // aus
                                    sString = oDevice.imageCell +
                                        '<td style="' + sError + '">' + sString + ' ' + _('set to Stage 2') + '</td>';
                                }
                                if (oAction.value == 255 && dss.staticDataModel.deviceHasThreeStages(oDevice)) { // aus
                                    sString = oDevice.imageCell +
                                        '<td style="' + sError + '">' + sString + ' ' + _('set to Stage 3') + '</td>';
                                }
                            }
                        } else {
                            sString = '<td colspan="2" class="h18" style="' + sError + '">' +
                                Math.round(oAction.value / 2.55) + '%' + ' ' + _('for') + ' ' + oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName +
                                '</td>';
                        }
                        sResult += sString;
                    }
                    break;
                }

                case DSS.ActionTypes.DEVICE_BLINK: { // dsuid
                    oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(oAction.dsuid), sError);

                    sResult += oDevice.imageCell +
                        '<td ' + (oDevice.imageCell ? '' : 'colspan="2" ') + ' class="h18" style="' + sError + '">' +
                        oDevice.deviceName + ' ' + _('in') + ' ' + oDevice.zoneName + ' ' + _('blinking') +
                        '</td>';
                    break;
                }

                case DSS.ActionTypes.ZONE_BLINK: { // group,zone
                    sString = '<td class="icon-small"><img src="' + dss.staticDataModel.baseImagePath + 'sceneIcons/scene_blink.png"></td>';
                    sString += '<td style="' + sError + '">' + dss.staticDataModel.getNameForGroup(oAction.zone, oAction.group);
                    if (oAction.zone != 0) {
                        sString += ' ' + _('in') + ' ' + dss.staticDataModel.getNameForZone(oAction.zone);
                    } else if (dss.staticDataModel.isClusterGroup(oAction.group)) {
                        // user groups have their own names
                    } else {
                        sString += ' ' + _('in all rooms');
                    }
                    sString += ' ' + _('blinking') + '</td>';
                    sResult += sString;
                    break;
                }

                case DSS.ActionTypes.CUSTOM_EVENT: {
                    sResult += '<td colspan="2" class="h18" style="' + sError + '">' + _('action') + ': ' + dss.staticDataModel.getHLENameFromId(oAction.event) + '</td>';
                    break;
                }

                case DSS.ActionTypes.URL: {
                    sResult += '<td colspan="2" class="h18" style="' + sError + '">URL: ' + oAction.url + '</td>';
                    break;
                }

                case DSS.ActionTypes.AUDIO_ACTIVITY: {
                    const zones = (dss.staticDataModel.getZoneArray() || [])
                        .reduce(function (zoneNamesById, zone) {
                            zoneNamesById[zone.ZoneID] = zone.name;
                            return zoneNamesById;
                        }, {});
                    const activityZones = JSON.parse(oAction.zones).map(function (zoneId) {
                        return zones[zoneId]
                            ? zones[zoneId]
                            : ' #' + zoneId;
                    }).join(', ');
                    dss.staticDataModel.getAudioActivities();
                    const activity = dss.staticDataModel.audioActivities.find(function (activity) {
                        return activity.id === oAction.contentSource;
                    });
                    const name = activity && activity.name || _('audio activity') + ' #' + oAction.contentSource;
                    sResult += '<td class="icon-small"><img src="' + dss.staticDataModel.baseImagePath + 'note.png"></td>' +
                        '<td colspan="2" class="h18" style="' + sError + '"> ' +
                        Ext.String.format2(_('{name} in zones {zones}'), {
                            name: name,
                            zones: activityZones
                        }) +
                        '</td>';
                    break;
                }
            }
            sResult += '</tr>';
        }

        return (view.up().isFieldActive(record)
            ? '<table>'
            : '<table style="color:grey;">') + sResult + '</table>';
    }
});

Ext.define('DSS.component.timespanColumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.timespancolumn',
    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        const hhmmss = secondsToHHMMSS(val);

        var fActive = view.up().isFieldActive(record);
        if (!fActive) {
            return '<span style="color:grey;">' + hhmmss + '</span>';
        } else {
            return hhmmss;
        }
    }
});

Ext.define('DSS.component.textColumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.textcolumn',
    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        var fActive = view.up().isFieldActive(record);
        if (!fActive) {
            return '<span style="color:grey;">' + Ext.String.htmlEncode(dss.escapeProperty(unescape(val))) + '</span>';
        }
        return Ext.String.htmlEncode(dss.escapeProperty(unescape(val)));
    }
});

Ext.define('DSS.component.deviceNameColumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.devicenamecolumn',
    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        var oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(val));
        var sString =
            '<tr>' +
                oDevice.imageCell +
                '<td>' + Ext.String.htmlEncode(oDevice.name ? oDevice.name : '?') + ' ' + _('in') + ' ' + oDevice.zoneName + '</td>' +
            '</tr>';

        return (view.up().isFieldActive(record)
            ? '<table>'
            : '<table style="color:grey;">' + sString + '</table>');
    }
});

Ext.define('DSS.component.deviceNameShortColumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.devicenameshortcolumn',
    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        var oDevice = Ext.parseDevice(dss.staticDataModel.getDevice(val));
        var sString =
            '<tr>' +
                oDevice.imageCell +
                '<td' + (oDevice.imageCell ? '' : ' colspan="2"') + '>' + Ext.String.htmlEncode(oDevice.name ? oDevice.name : '?') + '</td>' +
            '</tr>';

        return (view.up().isFieldActive(record)
            ? '<table>'
            : '<table style="color:grey;">') + sString + '</table>';
    }
});

Ext.define('DSS.component.deviceDSIDColumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.devicedsidcolumn',
    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        return (view.up().isFieldActive(record)
            ? '<span>' :
            '<span style="color:grey;">') + val + '</span>';
    }
});

Ext.define('DSS.component.dateColumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.datecolumn',
    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        return (view.up().isFieldActive(record)
            ? '<span>' :
            '<span style="color:grey;">') + Ext.Date.format(val, 'l, d.m.Y H:i') + '</span>';
    }
});


Ext.define('DSS.component.checkColumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.disabledcolumn',

    constructor: function () {
        this.addEvents('checkchange');
        this.callParent(arguments);
    },

    processEvent: function (type, view, cell, recordIndex) {    // , cellIndex, e
        if (type === 'mousedown') {
            var record = view.panel.store.getAt(recordIndex);

            var fActive = view.up().isFieldActive(record);
            if (!fActive) {
                return this.callParent(arguments);
            }

            fActive = view.up().isEnablerActive(record);
            if (!fActive) {
                return this.callParent(arguments);
            }

            var dataIndex = this.dataIndex;
            var checked = !record.get(dataIndex);

            record.set(dataIndex, checked);
            this.fireEvent('checkchange', this, record, recordIndex, checked);
            return this.callParent(arguments);
        }
    },

    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        var fActive = view.up().isFieldActive(record);
        if (!fActive) {
            return '<div style="margin:0 auto;width:13px;height:13px; background-position: 0px -13px;" />';
        }

        fActive = view.up().isEnablerActive(record);
        if (!fActive) {
            return '<div style="margin:0 auto;width:13px;height:13px; background-position: 0px -13px;" />';
        }

        const style = 'margin:0 auto;width:13px;height:13px;';
        return val
            ? '<div style="' + style + 'background-image:url(' + dss.staticDataModel.baseImagePath + 'dss/checkbox.gif);" />'
            : '<div style="' + style + 'background-image:url(' + dss.staticDataModel.baseImagePath + 'dss/checkbox.gif); background-position: 0px -13px;" />';
    }
});

Ext.define('DSS.component.toggledColumn', {
    extend: 'DSS.component.checkColumn',
    alias: 'widget.toggledcolumn',

    processEvent: function (type, view, cell, recordIndex) {    // , cellIndex, e
        if (type === 'mousedown') {
            const record = view.panel.store.getAt(recordIndex);
            const enabled = record.get(this.dataIndex) != null;
            if (!enabled) {
                return;
            }
        }
        return this.callParent(arguments);
    },

    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        const style = 'margin:0 auto;width:13px;height:13px;';
        var fActive = view.up().isFieldActive(record);
        if (!fActive) {
            return '<div style="' + style + ' background-position: 0px -13px;" />';
        }

        if (val == null) {
            return '<div style="' + style + 'background-image:url(' + dss.staticDataModel.baseImagePath + 'dss/checkbox.gif); background-position: -39px -39px;" />';
        }
        return val
            ? '<div style="' + style + 'background-image:url(' + dss.staticDataModel.baseImagePath + 'dss/checkbox.gif); background-position: 0px -13px;" />'
            : '<div style="' + style + 'background-image:url(' + dss.staticDataModel.baseImagePath + 'dss/checkbox.gif);" />';
    }
});


Ext.define('DSS.component.timeColumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.timecolumn',
    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        var sText;
        var fActive = view.up().isFieldActive(record);

        if (val.timeBase === 'daily') {
            sText = secondsToHHMM(val.offset) + ' ' + _('o\'clock');
            if (val.random != null) {
                sText = '&plusmn;' + sText;
            }
            return (fActive ? '<span>' : '<span style="color:grey;">') + sText + '</span>';
        } else {
            sText = secondsToHHMM(val.offset);

            if ([ 'dusk', 'dawn', 'sunrise', 'sunset' ].indexOf(val.timeBase) > -1) {
                if (val.offset == 0) {
                    sText = _('at ' + val.timeBase);
                } else if (val.offset < 0) {
                    sText = Ext.String.format2(_('{time} before ' + val.timeBase), { time: sText });
                } else {
                    sText = Ext.String.format2(_('{time} after ' + val.timeBase), { time: sText });
                }
            }
            if (val.limitBefore != null) {
                sText += Ext.String.format2(_('<br>but not past {hour}:{minute} o\'clock'), {
                    hour: secondsToHours(val.limitBefore),
                    minute: secondsToHourMinutes(val.limitBefore)
                });
            }
            if (val.limitAfter != null) {
                sText += Ext.String.format2(_('<br>but not before {hour}:{minute} o\'clock'), {
                    hour: secondsToHours(val.limitAfter),
                    minute: secondsToHourMinutes(val.limitAfter)
                });
            }
            if (val.random != null) {
                sText = '&plusmn;' + sText;
            }

            return (fActive ? '<span>' : '<span style="color:grey;">') + sText + '</span>';
        }
    }
});

Ext.define('DSS.component.recurrenceColumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.recurrencecolumn',
    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        var fActive = view.up().isFieldActive(record);
        var i;

        if (val.recurrenceBase === 'weekly') {
            var sString = '';

            let days = [];
            if (val.timeArray != null) {
                var recObject = val.timeArray;
                if (recObject) {
                    if (recObject.length === weekdaysCount.NONE) {
                        return (fActive ? '<span>' : '<span style="color:grey;">') + sString + _('no time defined') + '</span>';
                    }
                    if (recObject.length === weekdaysCount.ALL) {
                        return (fActive ? '<span>' : '<span style="color:grey;">') + sString + _('daily') + '</span>';
                    }

                    days = recObject.reduce(function (allDays, day) {
                        allDays[day] = true;
                        return allDays;
                    }, {});
                    if (recObject.length === weekdaysCount.WEEKEND && days.SA && days.SU) {
                        return (fActive ? '<span>' : '<span style="color:grey;">') + sString + _('on weekends') + '</span>';
                    }
                    if (recObject.length === weekdaysCount.WORKWEEK && days.MO && days.TU && days.WE && days.TH && days.FR) {
                        return (fActive ? '<span>' : '<span style="color:grey;">') + sString + _('MO-FR') + '</span>';
                    }
                    if (recObject.length === weekdaysCount.SINGLE) {
                        sString += Ext.Date.weekdayToWeekdayDescription(recObject[0]);
                        return (fActive ? '<span>' : '<span style="color:grey;">') + sString + '</span>';
                    }
                }
            } else {
                return (fActive ? '<span>' : '<span style="color:grey;">') + sString + _('no time defined') + '</span>';
            }

            const selectedDays = [];
            DSS.Weekdays.weekStartingFromMO().forEach(function (day) {
                selectedDays.push(days[day]
                    ? '<b>' + _(day) + '</b>'
                    : '<span style="color:#aaaaaa;text-decoration:line-through;">' + _(day) + '</span>');
            });
            return (fActive ? '<span>' : '<span style="color:grey;">') + sString + selectedDays.join(', ') + '</span>';

        } else {
            var sText = '';
            if (val.timeArray) {
                for (i = 0; i < val.timeArray.length; i++) {
                    var oDate = new Date(val.timeArray[i]);
                    if (!isNaN(oDate.getTime())) {
                        sText += '<tr><td>' + Ext.Date.format(oDate, 'l, d.m.Y') + '</td></tr>';
                    }
                }
            }
            return (fActive ? '<table>' : '<table style="color:grey;">') + sText + '</table>';
        }
    }
});

Ext.define('DSS.component.timeStampColumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.timestampcolumn',
    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        var pad = function (i) { return (i < 10 ? '0' + i : i); };

        if (val == null) {
            return _('-');
        } else if (!val) {
            return _('-');
        } else {
            var oRecord = val;
            var oDate = new Date();
            oDate.setTime(parseInt(oRecord, 10));
            var sText = pad(oDate.getDate()) + '.' + pad(oDate.getMonth() + 1) + '.' + oDate.getFullYear() + ' - ' + pad(oDate.getHours()) +
                    ':' + pad(oDate.getMinutes()) + ':' + pad(oDate.getSeconds());

            return (view.up().isFieldActive(record)
                ? '<span>'
                : '<span style="color:grey;">') + sText + '</span>';
        }
    }
});

Ext.define('DSS.component.EmailColumn', {
    extend: 'Ext.grid.column.Column',
    alias: 'widget.emailcolumn',
    renderer: function (val, metadata, record, rowIdx, colIdx, store, view) {
        const viewUpOutput = viewUpRendererOutput(val, metadata, record, rowIdx, colIdx, store, view);
        if (viewUpOutput != null) {
            return viewUpOutput;
        }

        var fActive = view.up().isFieldActive(record);
        var triggerObject = val;
        if (triggerObject.messageType === 'email') {
            var sSubject = Ext.String.htmlEncode(unescape(triggerObject.subject));
            var sMessage = Ext.String.htmlEncode(unescape(triggerObject.message));
            if (sSubject.length === 0) {
                sSubject = '<i>' + _('no subject') + '</i>';
            }
            if (sMessage.length === 0) {
                sMessage = '<i>' + _('no body') + '</i>';
            }
            var sText = '<tr>' +
                    '<td rowspan="3"><image src="' + dss.staticDataModel.baseImagePath + 'dss/email.png"></td>' +
                    '<td rowspan="3" style="width:5px;"><image src="' + dss.staticDataModel.baseImagePath + 't.gif"></td>' +
                    '<td>' + _('Receiver') + ':</td>' +
                    '<td>' + triggerObject.receiver + '</td></tr>' +
                '<tr><td>' + _('Subject') + ':</td><td>' + sSubject + '</td></tr>' +
                '<tr><td>' + _('Body') + ':</td><td>' + sMessage + '</td></tr>';
            return (fActive ? '<table>' : '<table style="color:grey;">') + sText + '</table>';
        } else {
            return (fActive ? '<span>' : '<span style="color:grey;">') + _('a message') + '</span>';
        }
    }
});


Ext.apply(Ext.data.SortTypes, {
    asZoneID: function (oValue) {
        return dss.staticDataModel.getNameForZone(oValue).toUpperCase();
    },
    asDSUID: function (oValue) {
        return dss.staticDataModel.getDevice(oValue).name.toUpperCase();
    },
    asTimestamp: function (oValue) {
        if (oValue == null) {
            return 0;
        } else if (!oValue) {
            return 0;
        } else {
            return parseInt(oValue, 10);
        }
    },
    asTimespan: function (oValue) {
        return oValue;
    },
    asTrigger: function (oValue) {
        var pad10 = function (i) {
            return i.toString().padStart(10, '0');
        };
        var sRetVal = '';
        var oAction = oValue[0];
        var oDevice;
        switch (oAction.type) {
            case 'zone-scene': { // scene,
                sRetVal = '0';
                sRetVal += pad10(oAction.group);
                sRetVal += pad10(oAction.scene);
                sRetVal += pad10(oAction.zone);
                break;
            }
            case 'device-scene': {
                sRetVal = '1';
                oDevice = dss.staticDataModel.getDevice(oAction.dsuid);
                sRetVal += pad10(oDevice.ZoneID);
                sRetVal += oAction.dsuid;
                sRetVal += pad10(oAction.scene);
                break;
            }
            case 'device-msg': {
                sRetVal = '2';
                oDevice = dss.staticDataModel.getDevice(oAction.dsuid);
                sRetVal += pad10(oDevice.ZoneID);
                sRetVal += oAction.dsuid;
                sRetVal += pad10(oAction.msg);
                break;
            }
            case 'custom-event': {
                sRetVal = '3';
                sRetVal += dss.staticDataModel.getHLENameFromId(oAction.event);
                break;
            }
            case 'device-sensor': {
                sRetVal = '4';
                oDevice = dss.staticDataModel.getDevice(oAction.dsuid);
                sRetVal += pad10(oDevice.ZoneID);
                sRetVal += oAction.dsuid;
                sRetVal += pad10(oAction.eventid);
                break;
            }
        }
        return sRetVal;
    },
    asBoolean: function (oValue) {
        return oValue ? 1 : 0;
    },
    asEmail: function (oValue) {
        return oValue.receiver + '_' + oValue.subject;
    },
    asAction: function (oValue) {
        var pad10 = function (i) {
            return i.toString().padStart(10, '0');
        };
        var oSelectValue = oValue[0];
        var sRet = '';
        var oDevice;
        switch (oSelectValue.type) {
            case 'zone-scene': {
                sRet = '0';
                sRet += pad10(oSelectValue.group);
                sRet += pad10(oSelectValue.scene);
                sRet += pad10(oSelectValue.zone);
                break;
            }
            case 'zone-blink': {
                sRet = '1';
                sRet += pad10(oSelectValue.group);
                sRet += pad10(oSelectValue.zone);
                break;
            }
            case 'device-scene': {
                sRet = '2';
                oDevice = dss.staticDataModel.getDevice(oSelectValue.dsuid);
                sRet += pad10(oDevice.ZoneID);
                sRet += oSelectValue.dsuid;
                sRet += pad10(oSelectValue.scene);
                break;
            }
            case 'device-blink': {
                sRet = '3';
                oDevice = dss.staticDataModel.getDevice(oSelectValue.dsuid);
                sRet += pad10(oDevice.ZoneID);
                sRet += oSelectValue.dsuid;
                break;
            }
            case 'device-value': {
                sRet = '4';
                oDevice = dss.staticDataModel.getDevice(oSelectValue.dsuid);
                sRet += pad10(oDevice.ZoneID);
                sRet += oSelectValue.dsuid;
                sRet += oSelectValue.value;
                break;
            }
            case 'custom-event': {
                sRet = '5';
                sRet += dss.staticDataModel.getHLENameFromId(oSelectValue.event);
                break;
            }
            case 'url': {
                sRet = '6';
                sRet += oSelectValue.url;
                break;
            }
        }
        return sRet;
    },
    asCondition: function (oValue) {
        var sRet = '';
        if (oValue.timeframe != null) {
            return '0';
        }
        if (oValue.weekdays != null) {
            return '1';
        }
        if (oValue.zoneState != null) {
            return '2';
        }
        if (oValue.systemState != null) {
            var oArray = [];
            for (var i = 0; i < oValue.systemState.length; i++) {
                oArray.push(oValue.systemState[i].name);
            }
            oArray = oArray.sort();
            return '3' + oArray[0];
        }
        return sRet;
    },
    asTime: function (oValue) {
        var retValue = 0;
        switch (oValue.timeBase) {
            case 'sunset':
                retValue = 350000;
                break;
            case 'sunrise':
                retValue = 250000;
                break;
            case 'dawn':
                retValue = 150000;
                break;
            default:
                retValue = 450000;
                break;
        }
        retValue += oValue.offset;
        return retValue;
    },
    asRecurrence: function (oValue) {
        if (oValue.timeArray != null) {
            var recObject = oValue.timeArray;
            if (recObject) {
                if (recObject.length === weekdaysCount.ALL) {
                    return 0;
                }
                const days = recObject.reduce(function (allDays, day) {
                    allDays[day] = true;
                    return allDays;
                }, {});
                if (recObject.length === weekdaysCount.WEEKEND && days.SA && days.SU) {
                    return 20;
                }
                if (recObject.length === weekdaysCount.WORKWEEK && days.MO && days.TU && days.WE && days.TH && days.FR) {
                    return 10;
                }

                if (recObject.length != weekdaysCount.SINGLE) {
                    var iReturnValue = 100;
                    if (days.MO) {
                        iReturnValue += 1;
                    }
                    if (days.TU) {
                        iReturnValue += 2;
                    }
                    if (days.WE) {
                        iReturnValue += 4;
                    }
                    if (days.TH) {
                        iReturnValue += 8;
                    }
                    if (days.FR) {
                        iReturnValue += 16;
                    }
                    if (days.SA) {
                        iReturnValue += 32;
                    }
                    if (days.SU) {
                        iReturnValue += 64;
                    }
                    return iReturnValue;
                } else {
                    if (days.MO) {
                        return 201;
                    }
                    if (days.TU) {
                        return 202;
                    }
                    if (days.WE) {
                        return 203;
                    }
                    if (days.TH) {
                        return 204;
                    }
                    if (days.FR) {
                        return 205;
                    }
                    if (days.SA) {
                        return 206;
                    }
                    if (days.SU) {
                        return 207;
                    }
                }
            }
        }
        return 1000;
    }
});

Ext.define('DSS.gridPanel', {
    extend: 'Ext.grid.Panel',
    config: {
        enableColumnHide: false,
        enableColumnMove: false,
        selType: 'rowmodel',
        viewConfig: { forceFit: true }
    },
    disabledProperty: null,
    store: null,
    editorWindow: null,
    curSelectedEvent: null,
    btnNew: null,
    btnEdit: null,
    btnDelete: null,
    btnDisable: null,
    btnEnable: null,
    btnTest: null,
    defaultSortCol: null,
    defaultSortDirection: 'ASC',
    fMultiselect: false,
    /* Parameter:
     *   * editor config
     *   * button config
     *   * columns
     *   * defaultSortCol
     *   * disabledProperty
     * */

    customColumnRenderer: {},
    customRendererOptions: {},
    statusBarVisible: true,
    statusBar: null,
    statusText: null,
    statusTextRight: null,
    grouperFn: null,
    inactiveField: null,
    inactiveEnablerField: null,
    setStatus: function (text, textRight) {
        this.statusText.setText(text);
        if (textRight != null) {
            this.statusTextRight.setText(textRight);
        }
    },
    isFieldActive: function (record) {
        if (this.inactiveField == null) {
            return true;
        }
        return !record.get(this.inactiveField);
    },
    isEnablerActive: function (record) {
        if (this.inactiveEnablerField == null) {
            return true;
        }
        return !record.get(this.inactiveEnablerField);
    },
    constructor: function (config) {
        var me = this;
        var i;
        if (config.multiselect != null) {
            this.fMultiselect = config.multiselect;
        }
        if (this.fMultiselect) {
            if (config.selModel == null) {
                config.selModel = {};
            }
            config.selModel.mode = 'MULTI';
        }
        if (config.grouperField != null) {
            this.grouperField = config.grouperField;
        }
        if (config.inactiveField != null) {
            this.inactiveField = config.inactiveField;
        }
        if (config.inactiveDisabledProperty != null) {
            this.inactiveEnablerField = config.inactiveDisabledProperty;
        }
        if (config.showStatusBar != null) {
            this.statusBarVisible = config.showStatusBar;
        }
        this.gridName = config.id || config.autoEl && config.autoEl['data-id'] || '';

        this.statusText = Ext.create('Ext.toolbar.TextItem', {
            autoEl: { 'data-id': this.gridName + 'StatusText' }
        });
        this.statusTextRight = Ext.create('Ext.toolbar.TextItem', {
            autoEl: { 'data-id': this.gridName + 'StatusTextRight' }
        });
        var refreshViewButton = Ext.create('Ext.button.Button', {
            text: _('Refresh'),
            handler: function () {
                me.setStatus(_('Refreshing...'));
                dss.staticDataModel.refreshDataModel();
            },
            icon: dss.staticDataModel.baseImagePath + 'buttonIcons/refresh.png'
        });
        addDataIdToElement(refreshViewButton, this.gridName + 'RefreshViewButton');
        this.statusBar = Ext.create('Ext.toolbar.Toolbar', {
            hidden: !this.statusBarVisible,
            items: [
                this.statusText,
                '->',
                refreshViewButton,
                this.statusTextRight
            ]
        });
        config.bbar = this.statusBar;

        this.initConfig(config);

        this.addEvents('loadDataStore');
        this.addEvents('saveEntry');
        this.addEvents('enableEntry');
        this.addEvents('disableEntry');
        this.addEvents('toggleColumn');
        this.addEvents('deleteEntry');
        this.addEvents('testEntry');
        this.addEvents('selectionChanged');

        if (config.disabledProperty != null) {
            this.disabledProperty = config.disabledProperty;
        }
        if (config.columns != null) {
            for (i = 0; i < config.columns.length; i++) {
                if (config.columns[i].xtype === 'disabledcolumn') {
                    config.columns[i].listeners = {
                        checkchange: function (column, oValue) {            // , dataIndex, checked
                            me.fireEvent(oValue.get('disabled') ? 'disableEntry' : 'enableEntry', me, oValue);
                        }
                    };
                }
                if (config.columns[i].xtype === 'toggledcolumn') {
                    config.columns[i].listeners = {
                        checkchange: function (column, oValue, dataIndex, checked) {
                            me.fireEvent('toggleColumn', me, oValue, column.dataIndex, checked);
                        }
                    };
                }
            }
        }

        var oStoreConfig = {
            storeId: config.storeConfig.storeId,
            fields: config.storeConfig.fields,
            proxy: {
                type: 'memory',
                reader: {
                    type: 'json',
                    root: 'items'
                }
            }
        };
        if (this.grouperField != null) {
            oStoreConfig.groupField = this.grouperField;
            config.features = [ { ftype: 'grouping', groupHeaderTpl: '{name}', enableGroupingMenu: false } ];
        }

        me.store = Ext.create('Ext.data.Store', oStoreConfig);

        if (config.scope != null) {
            me.store.filter([
                {
                    filterFn: function (item) {
                        if (config.scope.fieldID != null) {
                            if (config.scope.value != null) {
                                const fieldId = item.get(config.scope.fieldID);
                                return !fieldId || fieldId == config.scope.value;
                            }
                        }
                        return true;
                    }
                }
            ]);
        }

        if (config.buttonText.newButton != null) {
            this.btnNew = Ext.create('Ext.button.Button', {
                text: _(config.buttonText.newButton),
                iconCls: 'icon-add-action',
                handler: function () {
                    me.getSelectionModel().deselectAll();
                    me.editorWindow.show();
                    if (config.defaultNewEntryValues != null) {
                        me.editorWindow.setValues(Ext.JSON.decode(Ext.JSON.encode(config.defaultNewEntryValues)));
                    } else {
                        me.editorWindow.setValues(null);
                    }
                }
            });
            addDataIdToElement(this.btnNew, this.gridName + 'NewItemButton');
        }

        if (config.buttonText.editButton != null) {
            this.btnEdit = Ext.create('Ext.button.Button', {
                text: _(config.buttonText.editButton),
                iconCls: 'icon-edit-action',
                disabled: true,
                handler: function () {
                    me.editSelectedRecord();
                }
            });
            addDataIdToElement(this.btnEdit, this.gridName + 'EditItemButton');
        }

        if (config.buttonText.deleteButton != null) {
            this.btnDelete = Ext.create('Ext.button.Button', {
                text: _(config.buttonText.deleteButton),
                iconCls: 'icon-delete-action',
                disabled: true,
                handler: function () {
                    me.fireEvent('deleteEntry', me, me.curSelectedEvent);
                }
            });
            addDataIdToElement(this.btnDelete, this.gridName + 'DeleteItemButton');
        }

        if (this.disabledProperty != null) {
            if (config.buttonText.disableButton != null) {
                this.btnDisable = Ext.create('Ext.button.Button', {
                    text: _(config.buttonText.disableButton),
                    iconCls: 'icon-disable-action',
                    disabled: true,
                    handler: function () {
                        me.fireEvent('disableEntry', me, me.curSelectedEvent);
                    }
                });
                addDataIdToElement(this.btnDisable, this.gridName + 'DisableItemButton');
            }
            if (config.buttonText.enableButton != null) {
                this.btnEnable = Ext.create('Ext.button.Button', {
                    text: _(config.buttonText.enableButton),
                    iconCls: 'icon-enable-action',
                    hidden: true,
                    disabled: true,
                    handler: function () {
                        me.fireEvent('enableEntry', me, me.curSelectedEvent);
                    }
                });
                addDataIdToElement(this.btnEnable, this.gridName + 'EnableItemButton');
            }
        }
        if (config.buttonText.testButton != null) {
            this.btnTest = Ext.create('Ext.button.Button', {
                text: _(config.buttonText.testButton),
                iconCls: 'icon-test-action',
                disabled: true,
                handler: function () {
                    me.fireEvent('testEntry', me, me.curSelectedEvent);
                }
            });
            addDataIdToElement(this.btnTest, this.gridName + 'TestItemButton');
        }
        this.btnCustom = [];
        if (config.customButtons != null) {
            for (i = 0; i < config.customButtons.length; i++) {
                this.btnCustom.push(Ext.create('Ext.button.Button', {
                    text: _(config.customButtons[i].text),
                    iconCls: config.customButtons[i].iconCls,
                    disabled: false,
                    handler: function () {
                        this.thisFunction(me, me.curSelectedEvent);
                    }
                }));
                this.btnCustom[i].thisFunction = config.customButtons[i].handler;
                const btnId = config.customButtons[i].id
                    ? config.customButtons[i].id
                    : config.customButtons[i].autoEl && config.customButtons[i].autoEl['data-id']
                        ? config.customButtons[i].autoEl['data-id']
                        : i;
                addDataIdToElement(this.btnCustom[i], this.gridName + 'CustomButton' + btnId);
            }
        }
        this.btnCustomRAlign = [];
        if (config.customButtonsRAlign != null) {
            for (i = 0; i < config.customButtonsRAlign.length; i++) {
                this.btnCustomRAlign.push(Ext.create('Ext.button.Button', {
                    text: _(config.customButtonsRAlign[i].text),
                    iconCls: config.customButtonsRAlign[i].iconCls,
                    disabled: false,
                    handler: function () {
                        this.thisFunction(me, me.curSelectedEvent);
                    }
                }));
                this.btnCustomRAlign[i].thisFunction = config.customButtonsRAlign[i].handler;
                const btnId = config.customButtonsRAlign[i].id
                    ? config.customButtonsRAlign[i].id
                    : config.customButtonsRAlign[i].autoEl && config.customButtonsRAlign[i].autoEl['data-id']
                        ? config.customButtonsRAlign[i].autoEl['data-id']
                        : i;
                addDataIdToElement(this.btnCustomRAlign[i], this.gridName + 'CustomRAlignButton' + btnId);
            }
        }

        if (this.gridName) {
            config.editorConfig.id = this.gridName + 'Editor';
        }
        this.editorWindow = Ext.create('DSS.component.Editor', config.editorConfig);
        this.editorWindow.on({
            save: function (oValue) {
                me.fireEvent('saveEntry', me, oValue);
            }
        });

        var oPropertyArray = [];
        if (this.btnNew != null) {
            oPropertyArray.push(this.btnNew);
        }
        if (this.btnEdit != null) {
            oPropertyArray.push(this.btnEdit);
        }
        if (this.btnDisable != null) {
            oPropertyArray.push(this.btnDisable);
        }
        if (this.btnEnable != null) {
            oPropertyArray.push(this.btnEnable);
        }
        if (this.btnTest != null) {
            oPropertyArray.push(this.btnTest);
        }
        for (i = 0; i < this.btnCustom.length; i++) { oPropertyArray.push(this.btnCustom[i]); }
        oPropertyArray.push('->');
        for (i = 0; i < this.btnCustomRAlign.length; i++) {
            oPropertyArray.push(this.btnCustomRAlign[i]);
        }
        if (this.btnDelete != null) {
            oPropertyArray.push(this.btnDelete);
        }

        config.tbar = {
            items: oPropertyArray
        };

        this.callParent(arguments);

        this.customColumnRenderer = {};
        this.customRendererOptions = {};
        if (config.columns) {
            for (i = 0; i < config.columns.length; i++) {
                if (config.columns[i].customColumnRenderer) {
                    this.customColumnRenderer[i] = config.columns[i].customColumnRenderer;
                }
                if (config.columns[i].customRendererOptions) {
                    this.customRendererOptions[i] = config.columns[i].customRendererOptions;
                }
            }
        }

        this.on({
            selectionchange: function (view, selections) {  // , options
                if (me.inactiveField != null) {
                    if (selections) {
                        for (i = 0; i < selections.length; i++) {
                            if (selections[i].get(me.inactiveField)) {
                                view.deselect(selections[i]);
                                return;
                            }
                        }
                    }
                }
                if (selections.length) {
                    if (selections.length === 1) {
                        me.curSelectedEvent = selections[0];
                    } else {
                        me.curSelectedEvent = selections;
                    }

                    if (me.btnEdit != null) {
                        if (selections.length === 1) {
                            me.btnEdit.setDisabled(false);
                        } else {
                            me.btnEdit.setDisabled(true);
                        }
                    }
                    if (me.btnDelete != null) {
                        me.btnDelete.setDisabled(false);
                    }

                    if (selections.length === 1) {
                        if (me.disabledProperty != null) {
                            var fOverride = false;
                            if (me.inactiveEnablerField != null) {
                                fOverride = me.curSelectedEvent.get(me.inactiveEnablerField);
                            }

                            const disabled = me.curSelectedEvent.get('disabled');
                            if (!fOverride) {
                                if (me.btnDisable != null) {
                                    me.btnDisable.setDisabled(disabled);
                                    me.btnDisable.setVisible(!disabled);
                                }
                                if (me.btnEnable != null) {
                                    me.btnEnable.setDisabled(!disabled);
                                    me.btnEnable.setVisible(disabled);
                                }
                            } else {
                                if (me.btnDisable != null) {
                                    me.btnDisable.setDisabled(true);
                                    me.btnDisable.setVisible(false);
                                }
                                if (me.btnEnable != null) {
                                    me.btnEnable.setDisabled(true);
                                    me.btnEnable.setVisible(false);
                                }
                            }
                            if (me.btnTest != null) {
                                me.btnTest.setDisabled(disabled);
                            }
                        }
                    } else if (me.btnTest != null) {
                        me.btnTest.setDisabled(true);
                    }
                    if (selections.length > 1) {
                        if (me.btnEnable != null) {
                            me.btnEnable.setDisabled(false);
                            me.btnEnable.setVisible(true);
                        }
                        if (me.btnDisable != null) {
                            me.btnDisable.setDisabled(false);
                            me.btnDisable.setVisible(true);
                        }
                    }
                } else {
                    me.curSelectedEvent = null;
                    if (me.btnEdit != null) {
                        me.btnEdit.setDisabled(true);
                    }
                    if (me.btnDelete != null) {
                        me.btnDelete.setDisabled(true);
                    }
                    if (me.disabledProperty != null) {
                        if (me.btnDisable != null) {
                            me.btnDisable.setDisabled(true);
                            me.btnDisable.setVisible(true);
                        }
                        if (me.btnEnable != null) {
                            me.btnEnable.setDisabled(true);
                            me.btnEnable.setVisible(false);
                        }
                        if (me.btnTest != null) {
                            me.btnTest.setDisabled(true);
                        }
                    }
                }
                me.fireEvent('selectionChanged', me, selections);
            },
            itemdblclick: function () {
                if (me.curSelectedEvent != null) {
                    if (!me.btnEdit.disabled) {
                        me.editSelectedRecord();
                    }
                }
            },
            sortchange: function (header, col, direction) {
                me.defaultSortCol = col.dataIndex;
                me.defaultSortDirection = direction;
            }
        });
        var mThis = this;
        Ext.data.StoreManager.on('clear', function () {
            Ext.data.StoreManager.add(config.storeConfig.storeId, mThis.store);
            mThis.fireEvent('loadDataStore', mThis);
        });
        this.fireEvent('loadDataStore', this);
    },
    editSelectedRecord: function () {
        var oParams = {};
        if (this.curSelectedEvent && this.curSelectedEvent.fields && this.curSelectedEvent.fields.items) {
            for (var i = 0; i < this.curSelectedEvent.fields.items.length; i++) {
                var sKey = this.curSelectedEvent.fields.items[i].name;
                oParams[sKey] = this.curSelectedEvent.get(sKey);
            }
        }
        this.editorWindow.show();
        this.editorWindow.setValues(oParams);
    },
    loadParameter: function (values) {
        this.store.removeAll();
        this.store.add(values);
        this.store.sort(this.defaultSortCol, this.defaultSortDirection);
        if (this.curSelectedEvent != null) {
            this.getSelectionModel().select(this.curSelectedEvent);
        }
        this.store.filter();
        this.doLayout();
    }
});
